import six
from os import unlink
from collections import Counter
from itertools import chain
from hashlib import md5
from random import uniform

from Bio.Seq import translate
from Bio.Data.IUPACData import (
    ambiguous_dna_complement, ambiguous_rna_complement)

from dark.aa import AA_LETTERS
from dark.filter import TitleFilter
from dark.aa import PROPERTIES, PROPERTY_DETAILS, NONE
from dark.gor4 import GOR4


def _makeComplementTable(complementData):
    """
    Make a sequence complement table.

    @param complementData: A C{dict} whose keys and values are strings of
        length one. A key, value pair indicates a substitution that should
        be performed during complementation.
    @return: A 256 character string that can be used as a translation table
        by the C{translate} method of a Python string.
    """
    table = list(range(256))
    for _from, to in complementData.items():
        table[ord(_from[0])] = ord(to[0])
    return ''.join(map(chr, table))


class Read(object):
    """
    Hold information about a single read.

    @param id: A C{str} describing the read.
    @param sequence: A C{str} of sequence information (might be
        nucleotides or proteins).
    @param quality: An optional C{str} of phred quality scores. If not C{None},
        it must be the same length as C{sequence}.
    @raise ValueError: if the length of the quality string (if any) does not
        match the length of the sequence.
    """
    ALPHABET = None

    def __init__(self, id, sequence, quality=None):
        if quality is not None and len(quality) != len(sequence):
            raise ValueError(
                'Invalid read: sequence length (%d) != quality length (%d)' %
                (len(sequence), len(quality)))

        self.id = id
        self.sequence = sequence
        self.quality = quality

    def __eq__(self, other):
        return (self.id == other.id and
                self.sequence == other.sequence and
                self.quality == other.quality)

    def __ne__(self, other):
        return not self == other

    def __len__(self):
        return len(self.sequence)

    def __hash__(self):
        """
        Calculate a hash key for a read.

        @return: The C{int} hash key for the read.
        """
        if self.quality is None:
            return hash(md5(self.id.encode('UTF-8') + b'\0' +
                            self.sequence.encode('UTF-8')).digest())
        else:
            return hash(md5(self.id.encode('UTF-8') + b'\0' +
                            self.sequence.encode('UTF-8') + b'\0' +
                            self.quality.encode('UTF-8')).digest())

    def __getitem__(self, item):
        sequence = self.sequence[item]
        quality = None if self.quality is None else self.quality[item]
        return self.__class__(self.id, sequence, quality)

    def toString(self, format_):
        """
        Convert the read to a string format.

        @param format_: Either 'fasta', 'fastq' or 'fasta-ss'.
        @raise ValueError: if C{format_} is 'fastq' and the read has no quality
            information, if C{format_} is 'fasta-ss' and the read has no
            structure information, or if an unknown format is requested.
        @return: A C{str} representing the read in the requested format.
        """
        if format_ == 'fasta':
            return '>%s\n%s\n' % (self.id, self.sequence)
        elif format_ == 'fastq':
            if self.quality is None:
                raise ValueError('Read %r has no quality information' %
                                 self.id)
            else:
                return '@%s\n%s\n+%s\n%s\n' % (
                    self.id, self.sequence, self.id, self.quality)
        else:
            raise ValueError("Format must be either 'fasta', 'fastq' or "
                             "'fasta-ss'.")

    def toDict(self):
        """
        Get information about this read in a dictionary.

        @return: A C{dict} with keys/values for the attributes of self.
        """
        return {
            'id': self.id,
            'sequence': self.sequence,
            'quality': self.quality,
        }

    @classmethod
    def fromDict(cls, d):
        """
        Create a new instance from attribute values provided in a dictionary.

        @param d: A C{dict} with keys/values for the attributes of a new
            instance of this class. Keys 'id' and 'sequence' with C{str} values
            must be provided. A 'quality' C{str} key is optional.
        @return: A new instance of this class, with values taken from C{d}.
        """
        return cls(d['id'], d['sequence'], d.get('quality'))

    def lowComplexityFraction(self):
        """
        What fraction of a read's bases are in low-complexity regions?
        By convention, a region of low complexity is indicated by lowercase
        base letters.

        @return: The C{float} representing the fraction of bases in the
            read that are in regions of low complexity.
        """
        length = len(self)
        if length:
            lowerCount = len(list(filter(str.islower, self.sequence)))
            return float(lowerCount) / length
        else:
            return 0.0

    def walkHSP(self, hsp):
        """
        Provide information about exactly how a read matches a subject, as
        specified by C{hsp}.

        @return: A generator that yields (offset, residue, inMatch) tuples.
            The offset is the offset into the matched subject. The residue is
            the base in the read (which might be '-' to indicate a gap in the
            read was aligned with the subject at this offset). inMatch will be
            C{True} for residues that are part of the HSP match, and C{False}
            for the (possibly non-existent) parts of the read that fall outside
            the HSP (aka, the "whiskers" in an alignment graph).
        """

        # It will be easier to understand the following implementation if
        # you refer to the ASCII art illustration of an HSP in the
        # dark.hsp._Base class in hsp.py

        # Left whisker.
        readOffset = 0
        subjectOffset = hsp.readStartInSubject
        while subjectOffset < hsp.subjectStart:
            yield (subjectOffset, self.sequence[readOffset], False)
            readOffset += 1
            subjectOffset += 1

        # Match.
        for matchOffset, residue in enumerate(hsp.readMatchedSequence):
            yield (subjectOffset + matchOffset, residue, True)

        # Right whisker.
        readOffset = hsp.readEnd
        subjectOffset = hsp.subjectEnd
        while subjectOffset < hsp.readEndInSubject:
            yield (subjectOffset, self.sequence[readOffset], False)
            readOffset += 1
            subjectOffset += 1

    def checkAlphabet(self, count=10):
        """
        A function which checks whether the sequence in a L{dark.Read} object
        corresponds to its readClass. For AA reads, more testing is done in
        dark.Read.AARead.checkAlphabet.

        @param count: An C{int}, indicating how many bases or amino acids at
            the start of the sequence should be considered. If C{None}, all
            bases are checked.
        @return: C{True} if the alphabet characters in the first C{count}
            positions of sequence is a subset of the allowed alphabet for this
            read class, or if the read class has a C{None} alphabet.
        @raise ValueError: If the sequence alphabet is not a subset of the read
            class alphabet.
        """
        if count is None:
            readLetters = set(self.sequence.upper())
        else:
            readLetters = set(self.sequence.upper()[:count])
        # Check if readLetters is a subset of self.ALPHABET.
        if self.ALPHABET is None or readLetters.issubset(self.ALPHABET):
            return readLetters
        raise ValueError("Read alphabet (%r) is not a subset of expected "
                         "alphabet (%r) for read class %s." % (
                             ''.join(sorted(readLetters)),
                             ''.join(sorted(self.ALPHABET)),
                             str(self.__class__.__name__)))


class _NucleotideRead(Read):
    """
    Holds methods to work with nucleotide (DNA and RNA) sequences.
    """
    def translations(self):
        """
        Yield all six translations of a nucleotide sequence.

        @return: A generator that produces six L{TranslatedRead} instances.
        """
        rc = self.reverseComplement().sequence
        for reverseComplemented in False, True:
            for frame in 0, 1, 2:
                seq = rc if reverseComplemented else self.sequence
                # Get the suffix of the sequence for translation. I.e.,
                # skip 0, 1, or 2 initial bases, depending on the frame.
                # Note that this makes a copy of the sequence, which we can
                # then safely append 'N' bases to to adjust its length to
                # be zero mod 3.
                suffix = seq[frame:]
                lengthMod3 = len(suffix) % 3
                if lengthMod3:
                    suffix += ('NN' if lengthMod3 == 1 else 'N')
                yield TranslatedRead(self, translate(suffix), frame,
                                     reverseComplemented)

    def reverseComplement(self):
        """
        Reverse complement a nucleotide sequence.

        @return: The reverse complemented sequence as an instance of the
            current class.
        """
        quality = None if self.quality is None else self.quality[::-1]
        sequence = self.sequence.translate(self.COMPLEMENT_TABLE)[::-1]
        return self.__class__(self.id, sequence, quality)


class DNARead(_NucleotideRead):
    """
    Hold information and methods to work with DNA reads.
    """
    ALPHABET = set('ATCG')

    COMPLEMENT_TABLE = _makeComplementTable(ambiguous_dna_complement)


class RNARead(_NucleotideRead):
    """
    Hold information and methods to work with RNA reads.
    """
    ALPHABET = set('ATCGU')

    COMPLEMENT_TABLE = _makeComplementTable(ambiguous_rna_complement)


# Keep a single GOR4 instance that can be used by all AA reads. This saves us
# from re-scanning the GOR IV secondary structure database every time we make
# an AARead instance. This will be initialized when it's first needed.
_GOR4 = None


class AARead(Read):
    """
    Hold information and methods to work with AA reads.
    """
    ALPHABET = set(AA_LETTERS)

    def checkAlphabet(self, count=10):
        """
        A function which checks if an AA read really contains amino acids. This
        additional testing is needed, because the letters in the DNA alphabet
        are also in the AA alphabet.

        @param count: An C{int}, indicating how many bases or amino acids at
            the start of the sequence should be considered. If C{None}, all
            bases are checked.
        @return: C{True} if the alphabet characters in the first C{count}
            positions of sequence is a subset of the allowed alphabet for this
            read class, or if the read class has a C{None} alphabet.
        @raise ValueError: If a DNA sequence has been passed to AARead().
        """
        if six.PY3:
            readLetters = super().checkAlphabet(count)
        else:
            readLetters = Read.checkAlphabet(self, count)
        if len(self) > 10 and readLetters.issubset(set('ACGT')):
            raise ValueError('It looks like a DNA sequence has been passed to '
                             'AARead().')
        return readLetters

    def properties(self):
        """
        Translate an amino acid sequence to properties of the form:
        'F': HYDROPHOBIC | AROMATIC.

        @return: A generator yielding properties for the residues in the
            current sequence.
        """
        return (PROPERTIES.get(aa, NONE) for aa in self.sequence)

    def propertyDetails(self):
        """
        Translate an amino acid sequence to properties. Each property of the
        amino acid gets a value scaled from -1 to 1.

        @return: A list of property dictionaries.
        """
        return (PROPERTY_DETAILS.get(aa, NONE) for aa in self.sequence)

    def ORFs(self):
        """
        Find all ORFs in our sequence.

        @return: A generator that yields AAReadORF instances that correspond
            to the ORFs found in the AA sequence.
        """
        ORFStart = None
        openLeft = True
        seenStart = False

        for index, residue in enumerate(self.sequence):
            if residue == 'M':
                # Start codon.
                openLeft = False
                seenStart = True
            elif residue == '*':
                # Stop codon. Yield an ORF, if it has non-zero length.
                if ORFStart is not None and index - ORFStart > 0:
                    # The ORF has non-zero length.
                    yield AAReadORF(self, ORFStart, index, openLeft, False)
                    ORFStart = None
                # After a stop codon, we can no longer be open on the left
                # and we have no longer seen a start codon.
                openLeft = seenStart = False
            else:
                if (seenStart or openLeft) and ORFStart is None:
                    ORFStart = index

        # End of sequence. Yield the final ORF if there is one and it has
        # non-zero length.
        length = len(self.sequence)
        if ((seenStart or openLeft) and ORFStart is not None and
                length - ORFStart > 0):
            yield AAReadORF(self, ORFStart, length, openLeft, True)

    def gor4(self):
        """
        Get GOR IV secondary structure predictions (and the associated
        prediction probabilities).

        @return: A C{dict} with 'predictions' and 'probabilities' keys.
            The 'predictions' value is a C{str} of letters from {'H', 'E',
            'C'} for Helix, Beta Strand, Coil.  The probabilities value is
            a C{list} of C{float} triples, one for each amino acid in
            C{sequence}. The C{float} values are the probabilities assigned,
            in order, to Helix, Beta Strand, Coil.
        """
        # Initialize the GOR4 singleton, if necessary.
        global _GOR4
        if _GOR4 is None:
            _GOR4 = GOR4()

        return _GOR4.predict(self.sequence)


class AAReadWithX(AARead):
    """
    Hold information and methods to work with AA reads with additional
    characters.
    """
    ALPHABET = set(AA_LETTERS + ['X'])


class AAReadORF(AARead):
    """
    Hold information about an ORF from an AA read.

    @param originalRead: The original L{AARead} instance in which this ORF
        occurs.
    @param start: The C{int} offset where the ORF starts in the original read.
    @param stop: The Python-style C{int} offset of the end of the ORF in the
        original read. The final index is not included in the ORF.
    @param openLeft: A C{bool}. If C{True}, the ORF potentially begins before
        the sequence given in C{sequence}. I.e., the ORF-detection code started
        to examine a read assuming it was already in an ORF. If C{False}, a
        start codon was found preceeding this ORF.
    @param openRight: A C{bool}. If C{True}, the ORF potentially ends after
        the sequence given in C{sequence}. I.e., the ORF-detection code
        was in an ORF when it encountered the end of a read (so no stop codon
        was found). If C{False}, a stop codon was found in the read after this
        ORF.
    """
    def __init__(self, originalRead, start, stop, openLeft, openRight):
        if start < 0:
            raise ValueError('start offset (%d) less than zero' % start)
        if stop > len(originalRead):
            raise ValueError('stop offset (%d) > original read length (%d)' %
                             (stop, len(originalRead)))
        if start > stop:
            raise ValueError('start offset (%d) greater than stop offset (%d)'
                             % (start, stop))
        newId = '%s-%s%d:%d%s' % (originalRead.id,
                                  '(' if openLeft else '[',
                                  start, stop,
                                  ')' if openRight else ']')
        if originalRead.quality:
            AARead.__init__(self, newId, originalRead.sequence[start:stop],
                            originalRead.quality[start:stop])
        else:
            AARead.__init__(self, newId, originalRead.sequence[start:stop])
        self.start = start
        self.stop = stop
        self.openLeft = openLeft
        self.openRight = openRight

    def toDict(self):
        """
        Get information about this read in a dictionary.

        @return: A C{dict} with keys/values for the attributes of self.
        """
        if six.PY3:
            result = super().toDict()
        else:
            result = AARead.toDict(self)

        result.update({
            'start': self.start,
            'stop': self.stop,
            'openLeft': self.openLeft,
            'openRight': self.openRight,
        })

        return result

    @classmethod
    def fromDict(cls, d):
        """
        Create a new instance from attribute values provided in a dictionary.

        @param d: A C{dict} with keys/values for the attributes of a new
            instance of this class. Keys 'id' and 'sequence' with C{str} values
            must be provided. A 'quality' C{str} key is optional. Keys 'start'
            and 'stop' must have C{int} values. Keys 'openLeft' and 'openRight'
            are C{bool}, all keys are as described in the docstring for this
            class.
        @return: A new instance of this class, with values taken from C{d}.
        """
        # Make a dummy instance whose attributes we can set explicitly.
        new = cls(AARead('', ''), 0, 0, True, True)
        new.id = d['id']
        new.sequence = d['sequence']
        new.quality = d.get('quality')
        new.start = d['start']
        new.stop = d['stop']
        new.openLeft = d['openLeft']
        new.openRight = d['openRight']
        return new


class SSAARead(AARead):
    """
    Hold information to work with AAReads that have secondary structure
    information attached to them.

    Note that this class (currently) has no quality string associated with it.

    @param id: A C{str} describing the read.
    @param sequence: A C{str} of sequence information.
    @param structure: A C{str} of structure information.
    """
    def __init__(self, id, sequence, structure):
        if six.PY3:
            super().__init__(id, sequence)
        else:
            AARead.__init__(self, id, sequence)
        self.structure = structure

    def __eq__(self, other):
        return (self.id == other.id and
                self.sequence == other.sequence and
                self.structure == other.structure)

    def __hash__(self):
        """
        Calculate a hash key for a read.

        @return: The C{int} hash key for the read.
        """
        return hash(md5(self.id.encode('UTF-8') + b'\0' +
                        self.sequence.encode('UTF-8') + b'\0' +
                        self.structure.encode('UTF-8')).digest())

    def __getitem__(self, item):
        sequence = self.sequence[item]
        structure = None if self.structure is None else self.structure[item]
        return self.__class__(self.id, sequence, structure)

    def toString(self, format_='fasta-ss', structureSuffix=':structure'):
        """
        Convert the read to a string in PDB format (sequence & structure). This
        consists of two FASTA records, one for the sequence then one for the
        structure.

        @param format_: Either 'fasta-ss' or 'fasta'. In the former case, the
            structure information is returned. Otherwise, plain FASTA is
            returned.
        @param structureSuffix: The C{str} suffix to append to the read id
            for the second FASTA record, containing the structure information.
        @raise ValueError: If C{format_} is not 'fasta'.
        @return: A C{str} representing the read sequence and structure in PDB
            FASTA format.
        """
        if format_ == 'fasta-ss':
            return '>%s\n%s\n>%s%s\n%s\n' % (
                self.id, self.sequence, self.id, structureSuffix,
                self.structure)
        else:
            if six.PY3:
                return super().toString(format_=format_)
            else:
                return AARead.toString(self, format_=format_)

    def toDict(self):
        """
        Get information about this read in a dictionary.

        @return: A C{dict} with keys/values for the attributes of self.
        """
        return {
            'id': self.id,
            'sequence': self.sequence,
            'structure': self.structure,
        }

    @classmethod
    def fromDict(cls, d):
        """
        Create a new instance from attribute values provided in a dictionary.

        @param d: A C{dict} with keys/values for the attributes of a new
            instance of this class. Keys 'id', 'sequence', and 'structure'
            with C{str} values must be provided.
        @return: A new instance of this class, with values taken from C{d}.
        """
        return cls(d['id'], d['sequence'], d['structure'])


class SSAAReadWithX(SSAARead):
    """
    Hold information and methods to work with C{SSAARead}s allowing 'X'
    characters to appear in sequences.
    """
    ALPHABET = set(AA_LETTERS + ['X'])


class TranslatedRead(AARead):
    """
    Hold information about one DNA->AA translation of a Read.

    @param originalRead: The original DNA or RNA L{Read} instance from which
        this translation was obtained.
    @param sequence: The C{str} AA translated sequence.
    @param frame: The C{int} frame, either 0, 1, or 2.
    @param reverseComplemented: A C{bool}, C{True} if the original sequence
        must be reverse complemented to obtain this AA sequence.
    """
    def __init__(self, originalRead, sequence, frame,
                 reverseComplemented=False):
        if frame not in (0, 1, 2):
            raise ValueError('Frame must be 0, 1, or 2')
        newId = '%s-frame%d%s' % (originalRead.id, frame,
                                  'rc' if reverseComplemented else '')
        AARead.__init__(self, newId, sequence)
        self.frame = frame
        self.reverseComplemented = reverseComplemented

    def __eq__(self, other):
        return (AARead.__eq__(self, other) and
                self.frame == other.frame and
                self.reverseComplemented == other.reverseComplemented)

    def __ne__(self, other):
        return not self == other

    def toDict(self):
        """
        Get information about this read in a dictionary.

        @return: A C{dict} with keys/values for the attributes of self.
        """
        if six.PY3:
            result = super().toDict()
        else:
            result = AARead.toDict(self)

        result.update({
            'frame': self.frame,
            'reverseComplemented': self.reverseComplemented,
        })

        return result

    @classmethod
    def fromDict(cls, d):
        """
        Create a new instance from attribute values provided in a dictionary.

        @param d: A C{dict} with keys/values for the attributes of a new
            instance of this class. Keys 'id' and 'sequence' with C{str} values
            must be provided. A 'quality' C{str} key is optional. Key 'frame'
            must have an C{int} value. Key 'reverseComplemented' must be a
            C{bool}, all keys are as described in the docstring for this class.
        @return: A new instance of this class, with values taken from C{d}.
        """
        # Make a dummy instance whose attributes we can set explicitly.
        new = cls(AARead('', ''), 0, True)
        new.id = d['id']
        new.sequence = d['sequence']
        new.quality = d.get('quality')
        new.frame = d['frame']
        new.reverseComplemented = d['reverseComplemented']
        return new

    def maximumORFLength(self):
        """
        Return the length of the longest (possibly partial) ORF in a translated
        read. The ORF may originate or terminate outside the sequence, which is
        why the length is just a lower bound.
        """
        return max(len(orf) for orf in self.ORFs())


# Provide a mapping from all read class names to read classes. This can be
# useful in deserialization.
readClassNameToClass = {
    'AARead': AARead,
    'AAReadORF': AAReadORF,
    'AAReadWithX': AAReadWithX,
    'DNARead': DNARead,
    'RNARead': RNARead,
    'Read': Read,
    'SSAARead': SSAARead,
    'SSAAReadWithX': SSAAReadWithX,
    'TranslatedRead': TranslatedRead,
}


class Reads(object):
    """
    Maintain a collection of sequence reads.

    @param initialReads: If not C{None}, an iterable of C{Read} (or C{Read}
        subclass) instances.
    """

    def __init__(self, initialReads=None):
        self.additionalReads = []
        self._length = 0
        if initialReads is not None:
            list(map(self.add, initialReads))

    def add(self, read):
        """
        Add a read to this collection of reads.

        @param read: A C{Read} instance.
        """
        self.additionalReads.append(read)
        self._length += 1

    def __iter__(self):
        """
        Iterate through all the reads.

        As a side effect, calculate our length (the number of reads in this
        collection of reads).

        @return: A generator that yields reads. The returned read type depends
            on the kind of reads that were added to this instance.
        """
        # Set/reset self._length to 0 because __iter__ may be called more than
        # once.  If we don't set it to 0, a second call to __iter__ would
        # cause it to continue to be incremented from its original value.
        self._length = 0

        for read in chain(self.iter(), self.additionalReads):
            self._length += 1
            yield read

    def iter(self):
        """
        Placeholder to allow subclasses to provide reads.

        The idea here is that reads in this class and its subclasses can come
        from multiple sources. The first is reads that are added to the
        instance via the C{add} method. The (optional) second reads are those
        that are, for example, extracted from a file. E.g., the
        C{dark.reads.fasta.FastaReads} class (a subclass of C{Reads}) overrides
        this C{iter} method to provide reads from a file.
        """
        return []

    def __len__(self):
        # Note that __len__ reflects the number of reads that are currently
        # known.  If self.__iter__ has not been exhausted, we will not know
        # the true number of reads.
        return self._length

    def save(self, filename, format_='fasta'):
        """
        Write the reads to C{filename} in the requested format.

        @param filename: Either a C{str} file name to save into (the file will
            be overwritten) or an open file descriptor (e.g., sys.stdout).
        @param format_: A C{str} format to save as, either 'fasta', 'fastq' or
            'fasta-ss'.
        @raise ValueError: if C{format_} is 'fastq' and a read with no quality
            is present, or if an unknown format is requested.
        @return: C{self} in case our caller wants to chain the result into
            another method call.
        """
        format_ = format_.lower()

        if isinstance(filename, str):
            try:
                with open(filename, 'w') as fp:
                    for read in self:
                        fp.write(read.toString(format_))
            except ValueError:
                unlink(filename)
                raise
        else:
            # We have a file-like object.
            for read in self:
                filename.write(read.toString(format_))
        return self

    def filter(self, minLength=None, maxLength=None, removeGaps=False,
               whitelist=None, blacklist=None,
               titleRegex=None, negativeTitleRegex=None,
               truncateTitlesAfter=None, indices=None, head=None,
               removeDuplicates=False, modifier=None, randomSubset=None,
               trueLength=None, sampleFraction=None, sequenceNumbersFile=None):
        """
        Filter a set of reads to produce a matching subset.

        Note: there are many additional filtering options that could be added,
        e.g., on complexity fraction, on GC %, on quality, etc.

        @param minLength: The minimum acceptable length.
        @param maxLength: The maximum acceptable length.
        @param removeGaps: If C{True} remove all gaps ('-' characters) from the
            read sequences.
        @param whitelist: If not C{None}, a set of exact read ids that are
            always acceptable (though other characteristics, such as length,
            of a whitelisted id may rule it out).
        @param blacklist: If not C{None}, a set of exact read ids that are
            never acceptable.
        @param titleRegex: A regex that read ids must match.
        @param negativeTitleRegex: A regex that read ids must not match.
        @param truncateTitlesAfter: A string that read ids will be truncated
            beyond. If the truncated version of an id has already been seen,
            that sequence will be skipped.
        @param indices: Either C{None} or a set of C{int} indices corresponding
            to reads that are wanted. Indexing starts at zero.
        @param head: If not C{None}, the C{int} number of sequences at the
            start of the reads to return. Later sequences are skipped.
        @param removeDuplicates: If C{True} remove duplicated sequences.
        @param modifier: If not C{None}, a function that is passed a read
            and which either returns a read or C{None}. If it returns a read,
            that read is passed through the filter. If it returns C{None},
            the read is omitted. Such a function can be used to do customized
            filtering, to change sequence ids, etc.
        @param randomSubset: If not C{None}, an C{int} giving the number of
            sequences that should be returned. These will be selected at
            random, using an algorithm that does a single pass over the data
            and which needs to know (in advance) the total number of reads.
            Because a C{Reads} instance does not always know its length, it is
            possible to specify the number of reads by passing a C{trueLength}
            argument. Note that the random selection is done before any other
            filtering. Due to this, if you want to extract a random subset of
            the reads filtered in another way, it will be best to call filter
            twice rather than doing both types of filtering in one step. E.g.,
            you very likely should do this:
                reads.filter(maxLength=100).filter(randomSubset=20)
            rather than this:
                reads.filter(maxLength=100, randomSubset=20)
            The second version will extract a random subset of 20 reads and
            only return those that have length <= 100, so your result may have
            less than 20 reads. The former version extracts reads of the
            desired length and then takes 20 reads at random from that set, so
            you'll always get 20 raeds in your result, assuming there are at
            least that many reads satisfying the length filter.
        @param trueLength: The C{int} number of reads in this C{Reads}
            instance. Under normal circumstances it will not be necessary to
            pass this argument. However in some cases a subclass (e.g., with
            C{dark.fasta.FastaReads}) does not know its length until its data
            has been read from disk. In such cases, it is not possible to
            choose a random subset without keeping the subset in memory (which
            is undesirable). See
            https://en.wikipedia.org/wiki/Reservoir_sampling for one approach
            when the set size is unknown. However, it is possible to filter a
            random subset in a single pass over the data without keeping the
            set in memory if the set size is known. C{trueLength} makes it
            possible to pass the actual number of reads (this will obviously
            need to be known via some other mechanism). If C{None}, the length
            of the C{Reads} instance is used (but may be inaccurate in some
            cases, as above).
        @param sampleFraction: If not C{None}, a [0.0, 1.0] C{float} indicating
            a fraction of the reads that should be allowed to pass through the
            filter. The sample size will only be approximately the product of
            the C{sampleFraction} and the number of reads. The sample is taken
            at random. If you try to combine this filter with C{randomSubset}
            a C{ValueError} will be raised. If you need both filters, run them
            one after another.
        @param sequenceNumbersFile: If not C{None}, gives the C{str} name of a
            file containing (1-based) sequence numbers, in ascending order,
            one per line. Only those sequences matching the given numbers will
            be kept.
        @raises ValueError: If C{randomSubset} and C{sampleFraction} are both
            specified or if the sequence numbers in C{sequenceNumbersFile} are
            non-positive or not ascending.
        @return: A new C{Reads} instance, with reads filtered as requested.
        """

        # TODO, when/if needed: make it possible to pass a seed for the RNG
        # when randomSubset or sampleFraction are used. Also possible is to
        # save and restore the state of the RNG and/or to optionally add
        # 'seed=XXX' to the end of the id of the first read, etc.

        return Reads(self._filter(
            minLength=minLength, maxLength=maxLength, removeGaps=removeGaps,
            whitelist=whitelist, blacklist=blacklist, titleRegex=titleRegex,
            negativeTitleRegex=negativeTitleRegex,
            truncateTitlesAfter=truncateTitlesAfter, indices=indices,
            head=head, removeDuplicates=removeDuplicates, modifier=modifier,
            randomSubset=randomSubset, trueLength=trueLength,
            sampleFraction=sampleFraction,
            sequenceNumbersFile=sequenceNumbersFile))

    def _filter(self, minLength=None, maxLength=None, removeGaps=False,
                whitelist=None, blacklist=None,
                titleRegex=None, negativeTitleRegex=None,
                truncateTitlesAfter=None, indices=None, head=None,
                removeDuplicates=False, modifier=None, randomSubset=None,
                trueLength=None, sampleFraction=None,
                sequenceNumbersFile=None):
        """
        Filter a set of reads to produce a matching subset.

        See docstring for self.filter (above) for parameter docs.

        @return: A generator that yields C{Read} instances.
        """

        def _wantedSequences(filename):
            """
            Read and yield integer sequence numbers from a file.

            @raise ValueError: If the sequence numbers are not all positive or
                are not ascending.
            @return: A generator that yields C{int} sequence numbers.
            """
            with open(filename) as fp:
                lastNumber = None
                for line in fp:
                    n = int(line)
                    if lastNumber is None:
                        if n < 1:
                            raise ValueError(
                                'First line of sequence number file %r must '
                                'be at least 1.' % filename)
                        lastNumber = n
                        yield n
                    else:
                        if n > lastNumber:
                            lastNumber = n
                            yield n
                        else:
                            raise ValueError(
                                'Line number file %r contains non-ascending '
                                'numbers %d and %d.' %
                                (filename, lastNumber, n))

        if randomSubset is not None and sampleFraction is not None:
            raise ValueError('randomSubset and sampleFraction cannot be '
                             'used simultaneously in a filter. Call filter '
                             'twice instead.')

        if sequenceNumbersFile is None:
            nextWantedSequenceNumber = None
            wantedSequenceNumberGeneratorExhausted = False
        else:
            wantedSequenceNumerGenerator = _wantedSequences(
                sequenceNumbersFile)
            try:
                nextWantedSequenceNumber = next(wantedSequenceNumerGenerator)
            except StopIteration:
                # There was a sequence number file, but it was empty.
                return
            else:
                wantedSequenceNumberGeneratorExhausted = False

        if (whitelist or blacklist or titleRegex or negativeTitleRegex or
                truncateTitlesAfter):
            titleFilter = TitleFilter(
                whitelist=whitelist, blacklist=blacklist,
                positiveRegex=titleRegex, negativeRegex=negativeTitleRegex,
                truncateAfter=truncateTitlesAfter)
        else:
            titleFilter = None

        if removeDuplicates:
            sequencesSeen = set()

        if sampleFraction is not None:
            if sampleFraction == 0.0:
                # The filter returns nothing.
                return
            elif sampleFraction == 1.0:
                # Passing 1.0 can be treated the same as passing no value.
                # This makes the loop code simpler.
                sampleFraction = None

        if randomSubset is not None and trueLength is None:
            trueLength = self._length

        yieldCount = 0

        for readIndex, read in enumerate(self):

            if wantedSequenceNumberGeneratorExhausted:
                return

            if nextWantedSequenceNumber is not None:
                if readIndex + 1 == nextWantedSequenceNumber:
                    # We want this sequence.
                    try:
                        nextWantedSequenceNumber = next(
                            wantedSequenceNumerGenerator)
                    except StopIteration:
                        # The sequence number iterator ran out of sequence
                        # numbers.  We must let the rest of the filtering
                        # continue for the current sequence in case we
                        # throw it out for other reasons (as we might have
                        # done for any of the earlier wanted sequence
                        # numbers).
                        wantedSequenceNumberGeneratorExhausted = True
                else:
                    # This sequence isn't wanted.
                    continue

            if (sampleFraction is not None and
                    uniform(0.0, 1.0) > sampleFraction):
                # Note that we don't have to worry about the 0.0 or 1.0
                # cases in the above if, as they have been dealt with
                # before the loop.
                continue

            if randomSubset is not None:
                if yieldCount == randomSubset:
                    # The random subset has already been fully returned.
                    # There's no point in going any further through the input.
                    return
                elif uniform(0.0, 1.0) > ((randomSubset - yieldCount) /
                                          (trueLength - readIndex)):
                    continue

            if head is not None and readIndex == head:
                # We're completely done.
                return

            readLen = len(read)
            if ((minLength is not None and readLen < minLength) or
                    (maxLength is not None and readLen > maxLength)):
                continue

            if removeGaps:
                sequence = read.sequence.replace('-', '')
                read = read.__class__(read.id, sequence, read.quality)

            if (titleFilter and
                    titleFilter.accept(read.id) == TitleFilter.REJECT):
                continue

            if indices is not None and readIndex not in indices:
                continue

            if removeDuplicates:
                if read.sequence in sequencesSeen:
                    continue
                sequencesSeen.add(read.sequence)

            if modifier:
                modified = modifier(read)
                if modified is None:
                    continue
                else:
                    read = modified

            yield read
            yieldCount += 1

    def summarizePosition(self, index):
        """
        Compute residue counts a specific sequence index.

        @param index: an C{int} index into the sequence.
        @return: A C{dict} with the count of too-short (excluded) sequences,
            and a Counter instance giving the residue counts.
        """
        countAtPosition = Counter()
        excludedCount = 0

        for read in self:
            try:
                countAtPosition[read.sequence[index]] += 1
            except IndexError:
                excludedCount += 1

        return {
            'excludedCount': excludedCount,
            'countAtPosition': countAtPosition
        }



from unittest import TestCase
from six import StringIO, assertRaisesRegex
from six.moves import builtins
from contextlib import contextmanager

from dark.proteins import (
    ProteinGrouper, VirusSampleFASTA, readSampleTotalReadCounts)

try:
    from unittest.mock import patch
except ImportError:
    from mock import patch

from .mocking import File


class TestReadSampleTotalReadCounts(TestCase):
    """
    Tests for the readSampleTotalReadCounts function.
    """

    def testNonIntegerCount(self):
        """
        """
        error = "^invalid literal for int\(\) with base 10: 'xxx\\\\n'$"
        fp = StringIO('sample-1 xxx\n')
        assertRaisesRegex(self, ValueError, error, readSampleTotalReadCounts,
                          fp)

    def testEmpty(self):
        """
        If the total read count file is empty, an empty dict must be
        returned.
        """
        fp = StringIO()
        self.assertEqual({}, readSampleTotalReadCounts(fp))

    def testNotEmpty(self):
        """
        If the total read count file is not empty, the expected dict must be
        returned.
        """
        fp = StringIO('sample-1 322\nsample-2 345\n')
        self.assertEqual(
            {
                'sample-1': 322,
                'sample-2': 345,
            },
            readSampleTotalReadCounts(fp))


class TestProteinGrouper(TestCase):
    """
    Tests for the dark.proteins.ProteinGrouper class
    """

    def testNoAssetDir(self):
        """
        If no asset directorey is given to a protein grouper, its _assetDir
        attribute be the default ('out').
        """
        pg = ProteinGrouper()
        self.assertEqual('out', pg._assetDir)

    def testAssetDir(self):
        """
        If an asset directorey is given to a protein grouper, its _assetDir
        attribute be set to hold that value.
        """
        pg = ProteinGrouper(assetDir='xxx')
        self.assertEqual('xxx', pg._assetDir)

    def testNoRegex(self):
        """
        If no regex is given to a protein grouper, its _sampleNameRegex
        attribute be None.
        """
        pg = ProteinGrouper()
        self.assertEqual(None, pg._sampleNameRegex)

    def testNoFiles(self):
        """
        If no files have been given to a protein grouper, its sample names and
        virus titles attributes must both be empty.
        """
        pg = ProteinGrouper()
        self.assertEqual({}, pg.virusTitles)
        self.assertEqual({}, pg.sampleNames)

    def testOneLineInOneFile(self):
        """
        If a protein grouper is given one file with one line, its virusTitles
        dict must be as expected.
        """
        fp = StringIO(
            '0.77 46.6 48.1 5 6 74 gi|327|X|I44.6 ubiquitin [Lausannevirus]\n')
        pg = ProteinGrouper()
        pg.addFile('sample-filename', fp)
        self.assertEqual(
            {
                'Lausannevirus': {
                    'sample-filename': {
                        'proteins': [
                            {
                                'bestScore': 48.1,
                                'bluePlotFilename': 'out/0.png',
                                'coverage': 0.77,
                                'fastaFilename': 'out/0.fasta',
                                'hspCount': 6,
                                'index': 0,
                                'medianScore': 46.6,
                                'outDir': 'out',
                                'proteinLength': 74,
                                'proteinTitle': 'gi|327|X|I44.6 ubiquitin',
                                'proteinURL': (
                                    'http://www.ncbi.nlm.nih.gov/nuccore/I44'),
                                'readCount': 5,
                            },
                        ],
                        'uniqueReadCount': None,
                    },
                }
            },
            pg.virusTitles)

    def testOneLineInOneFileTitle(self):
        """
        If a protein grouper is given one file with one line, its _title method
        must return the expected string.
        """
        fp = StringIO(
            '0.77 46.6 48.1 5 6 74 gi|327|X|I44.6 ubiquitin [Lausannevirus]\n')
        pg = ProteinGrouper()
        pg.addFile('sample-filename', fp)
        self.assertEqual('1 virus found in 1 sample', pg._title())

    def testTwoLinesInOneFileTitle(self):
        """
        If a protein grouper is given one file with two protein lines, each
        from a different virus, its _title method must return the expected
        string.
        """
        fp = StringIO(
            '0.77 46.6 48.1 5 6 74 gi|327|X|I44.6 ubiquitin [Lausannevirus]\n'
            '0.77 46.6 48.1 5 6 74 gi|327|X|I44.6 ubiquitin [X Virus]\n'
            )
        pg = ProteinGrouper()
        pg.addFile('sample-filename', fp)
        self.assertEqual('2 viruses found in 1 sample', pg._title())

    def testTwoLinesInOneFileSameVirus(self):
        """
        If a protein grouper is given one file with two lines from the same
        virus, its virusTitles dict must be as expected.
        """
        fp = StringIO(
            '0.63 41.3 44.2 9 9 12 gi|327410| protein 77 [Lausannevirus]\n'
            '0.77 46.6 48.1 5 6 74 gi|327409| ubiquitin [Lausannevirus]\n'
        )
        pg = ProteinGrouper()
        pg.addFile('sample-filename', fp)
        self.assertEqual(
            {
                'Lausannevirus': {
                    'sample-filename': {
                        'proteins': [
                            {
                                'bestScore': 44.2,
                                'bluePlotFilename': 'out/0.png',
                                'coverage': 0.63,
                                'fastaFilename': 'out/0.fasta',
                                'hspCount': 9,
                                'index': 0,
                                'medianScore': 41.3,
                                'outDir': 'out',
                                'proteinLength': 12,
                                'proteinTitle': 'gi|327410| protein 77',
                                'proteinURL': None,
                                'readCount': 9,
                            },
                            {
                                'bestScore': 48.1,
                                'bluePlotFilename': 'out/1.png',
                                'coverage': 0.77,
                                'fastaFilename': 'out/1.fasta',
                                'hspCount': 6,
                                'index': 1,
                                'medianScore': 46.6,
                                'outDir': 'out',
                                'proteinLength': 74,
                                'proteinTitle': 'gi|327409| ubiquitin',
                                'proteinURL': None,
                                'readCount': 5,
                            },
                        ],
                        'uniqueReadCount': None,
                    },
                },
            },
            pg.virusTitles)

    def testTwoLinesInOneFileDifferentViruses(self):
        """
        If a protein grouper is given one file with two lines from different
        viruss, its virusTitles dict must be as expected.
        """
        fp = StringIO(
            '0.63 41.3 44.2 9 9 12 gi|327410| protein 77 [Lausannevirus]\n'
            '0.77 46.6 48.1 5 6 74 gi|327409| ubiquitin [Hepatitis B virus]\n'
        )
        pg = ProteinGrouper()
        pg.addFile('sample-filename', fp)
        self.assertEqual(
            {
                'Lausannevirus': {
                    'sample-filename': {
                        'proteins': [
                            {
                                'bestScore': 44.2,
                                'bluePlotFilename': 'out/0.png',
                                'coverage': 0.63,
                                'fastaFilename': 'out/0.fasta',
                                'hspCount': 9,
                                'index': 0,
                                'medianScore': 41.3,
                                'outDir': 'out',
                                'proteinLength': 12,
                                'proteinTitle': 'gi|327410| protein 77',
                                'proteinURL': None,
                                'readCount': 9,
                            },
                        ],
                        'uniqueReadCount': None,
                    },
                },
                'Hepatitis B virus': {
                    'sample-filename': {
                        'proteins': [
                            {
                                'bestScore': 48.1,
                                'bluePlotFilename': 'out/1.png',
                                'coverage': 0.77,
                                'fastaFilename': 'out/1.fasta',
                                'hspCount': 6,
                                'index': 1,
                                'medianScore': 46.6,
                                'outDir': 'out',
                                'proteinLength': 74,
                                'proteinTitle': 'gi|327409| ubiquitin',
                                'proteinURL': None,
                                'readCount': 5,
                            },
                        ],
                        'uniqueReadCount': None,
                    },
                },
            },
            pg.virusTitles)

    def testOneLineInEachOfTwoFilesSameVirus(self):
        """
        If a protein grouper is given two files, each with one line from the
        same virus, its virusTitles dict must be as expected.
        """
        fp1 = StringIO(
            '0.63 41.3 44.2 9 9 12 gi|327410| protein 77 [Lausannevirus]\n'
        )
        fp2 = StringIO(
            '0.77 46.6 48.1 5 6 74 gi|327409| ubiquitin [Lausannevirus]\n'
        )
        pg = ProteinGrouper()
        pg.addFile('sample-filename-1', fp1)
        pg.addFile('sample-filename-2', fp2)
        self.assertEqual(
            {
                'Lausannevirus': {
                    'sample-filename-1': {
                        'proteins': [
                            {
                                'bestScore': 44.2,
                                'bluePlotFilename': 'out/0.png',
                                'coverage': 0.63,
                                'fastaFilename': 'out/0.fasta',
                                'hspCount': 9,
                                'index': 0,
                                'medianScore': 41.3,
                                'outDir': 'out',
                                'proteinLength': 12,
                                'proteinTitle': 'gi|327410| protein 77',
                                'proteinURL': None,
                                'readCount': 9,
                            },
                        ],
                        'uniqueReadCount': None,
                    },
                    'sample-filename-2': {
                        'proteins': [
                            {
                                'bestScore': 48.1,
                                'bluePlotFilename': 'out/0.png',
                                'coverage': 0.77,
                                'fastaFilename': 'out/0.fasta',
                                'hspCount': 6,
                                'index': 0,
                                'medianScore': 46.6,
                                'outDir': 'out',
                                'proteinLength': 74,
                                'proteinTitle': 'gi|327409| ubiquitin',
                                'proteinURL': None,
                                'readCount': 5,
                            },
                        ],
                        'uniqueReadCount': None,
                    },
                },
            },
            pg.virusTitles)

    def testOneLineInEachOfTwoFilesSameVirusTitle(self):
        """
        If a protein grouper is given two files, each with one line from the
        same virus, its _title method must return the expected string.
        """
        fp1 = StringIO(
            '0.63 41.3 44.2 9 9 12 gi|327410| protein 77 [Lausannevirus]\n'
        )
        fp2 = StringIO(
            '0.77 46.6 48.1 5 6 74 gi|327409| ubiquitin [Lausannevirus]\n'
        )
        pg = ProteinGrouper()
        pg.addFile('sample-filename-1', fp1)
        pg.addFile('sample-filename-2', fp2)
        self.assertEqual('1 virus found in 2 samples', pg._title())

    def testOneLineInEachOfTwoFilesDifferentViruses(self):
        """
        If a protein grouper is given two files in two different directories,
        each with one line from the different viruses, its virusTitles dict
        must be as expected.
        """
        fp1 = StringIO(
            '0.63 41.3 44.2 9 9 12 gi|327410| protein 77 [Lausannevirus]\n'
        )
        fp2 = StringIO(
            '0.77 46.6 48.1 5 6 74 gi|327409| ubiquitin [Hepatitis B virus]\n'
        )
        pg = ProteinGrouper()
        pg.addFile('dir-1/sample-filename-1', fp1)
        pg.addFile('dir-2/sample-filename-2', fp2)
        self.assertEqual(
            {
                'Lausannevirus': {
                    'dir-1/sample-filename-1': {
                        'proteins': [
                            {
                                'bestScore': 44.2,
                                'bluePlotFilename': 'dir-1/out/0.png',
                                'coverage': 0.63,
                                'fastaFilename': 'dir-1/out/0.fasta',
                                'hspCount': 9,
                                'index': 0,
                                'medianScore': 41.3,
                                'outDir': 'dir-1/out',
                                'proteinLength': 12,
                                'proteinTitle': 'gi|327410| protein 77',
                                'proteinURL': None,
                                'readCount': 9,
                            },
                        ],
                        'uniqueReadCount': None,
                    },
                },
                'Hepatitis B virus': {
                    'dir-2/sample-filename-2': {
                        'proteins': [
                            {
                                'bestScore': 48.1,
                                'bluePlotFilename': 'dir-2/out/0.png',
                                'coverage': 0.77,
                                'fastaFilename': 'dir-2/out/0.fasta',
                                'hspCount': 6,
                                'index': 0,
                                'medianScore': 46.6,
                                'outDir': 'dir-2/out',
                                'proteinLength': 74,
                                'proteinTitle': 'gi|327409| ubiquitin',
                                'proteinURL': None,
                                'readCount': 5,
                            },
                        ],
                        'uniqueReadCount': None,
                    },
                },
            },
            pg.virusTitles)

    def testOneLineInEachOfTwoFilesDifferentVirusesTitle(self):
        """
        If a protein grouper is given two files, each with one line from
        different viruses, its _title method must return the expected string.
        """
        fp1 = StringIO(
            '0.63 41.3 44.2 9 9 12 gi|327410| protein 77 [Lausannevirus]\n'
        )
        fp2 = StringIO(
            '0.77 46.6 48.1 5 6 74 gi|327409| ubiquitin [HBV]\n'
        )
        pg = ProteinGrouper()
        pg.addFile('sample-filename-1', fp1)
        pg.addFile('sample-filename-2', fp2)
        self.assertEqual('2 viruses found in 2 samples', pg._title())

    def testNoFilesToStr(self):
        """
        If no files have been given to a protein grouper, its text string
        format must as expected.
        """
        pg = ProteinGrouper()
        self.assertEqual('0 viruses found in 0 samples\n', pg.toStr())

    def testNoFilesToHTML(self):
        """
        If no files have been given to a protein grouper, its HTML string
        format must as expected.
        """
        pg = ProteinGrouper()
        self.assertEqual(
            '\n'.join([
                '<html>',
                '<head>',
                '<title>',
                '0 viruses found in 0 samples',
                '</title>',
                '</head>',
                '<body>',
                '<style>',
                '            body {',
                '                margin-left: 2%;',
                '                margin-right: 2%;',
                '            }',
                '            .sample {',
                '                margin-bottom: 2px;',
                '            }',
                '            .sample-name {',
                '                color: red;',
                '            }',
                '            .index {',
                '                font-size: small;',
                '            }',
                '            .protein-title {',
                '                font-family: "Courier New", Courier, '
                'monospace;',
                '            }',
                '            .stats {',
                '                font-family: "Courier New", Courier, '
                'monospace;',
                '                white-space: pre;',
                '            }',
                '            .protein-list {',
                '                margin-top: 2px;',
                '            }',
                '</style>',
                '</head>',
                '<body>',
                '<h1>0 viruses found in 0 samples</h1>',
                '<h2>Virus index</h2>',
                '</p>',
                '<h2>Sample index</h2>',
                '</p>',
                '<h1>Viruses by sample</h1>',
                '<p>',
                'In the bullet point protein lists below, there are eight '
                'fields:',
                '<ol>',
                '<li>Coverage fraction.</li>',
                '<li>Median bit score.</li>',
                '<li>Best bit score.</li>',
                '<li>Read count.</li>',
                '<li>HSP count (a read can match a protein more than once).'
                '</li>',
                '<li>Protein length (in AAs).</li>',
                '<li>Index (just ignore this).</li>',
                '<li>Protein name.</li>',
                '</ol>',
                '</p>',
                '<h1>Samples by virus</h1>',
                '<p>',
                'In the bullet point protein lists below, there are eight '
                'fields:',
                '<ol>',
                '<li>Coverage fraction.</li>',
                '<li>Median bit score.</li>',
                '<li>Best bit score.</li>',
                '<li>Read count.</li>',
                '<li>HSP count (a read can match a protein more than once).'
                '</li>',
                '<li>Protein length (in AAs).</li>',
                '<li>Index (just ignore this).</li>',
                '<li>Protein name.</li>',
                '</ol>',
                '</p>',
                '</body>',
                '</html>',
                ]),
            pg.toHTML())

    def testOneLineInOneFileToStr(self):
        """
        If a protein grouper is given one file with one line, its toStr method
        must produce the expected result.
        """
        fp = StringIO(
            '0.77 46.6 48.1 5 6 74 gi|32|X|I4 protein X [HBV]\n')
        pg = ProteinGrouper()
        pg.addFile('sample-filename', fp)
        self.assertEqual(
            '1 virus found in 1 sample\n'
            '\n'
            'HBV (in 1 sample)\n'
            '  sample-filename (1 protein, 5 reads)\n'
            '    0.77\t46.60\t48.10\t   5\t   6\t  0\tgi|32|X|I4 protein X\n',
            pg.toStr())


class TestVirusSampleFASTA(TestCase):
    """
    Tests for the VirusSampleFASTA class.
    """

    def testOpenNotCalledOnRepeatedCall(self):
        """
        If a repeated call to virusSampleFASTA.add is made with the same
        arguments, no file should be read because the original result value is
        cached.
        """
        class Open(object):
            def __init__(self, test, manager):
                self.test = test
                self.manager = manager
                self.count = 0

            def sideEffect(self, filename, *args, **kwargs):
                if self.count == 0:
                    self.test.assertEqual('out/0.fasta', filename)
                    self.count += 1
                    return File(['>id1\n', 'ACTG\n'])
                elif self.count == 1:
                    self.test.assertEqual('out/virus-0-sample-0.fasta',
                                          filename)
                    self.count += 1
                    return self.manager
                else:
                    self.test.fail(
                        'We are only supposed to be called twice. '
                        'Filename: %r, Args: %r, Keyword args: %r.' %
                        (filename, args, kwargs))

        fp = StringIO(
            '0.63 41.3 44.2 9 9 12 gi|327410| protein 77 [Lausannevirus]\n'
        )
        fastaIO = StringIO()

        @contextmanager
        def manager():
            yield fastaIO

        pg = ProteinGrouper()
        pg.addFile('filename-1', fp)
        virusSampleFASTA = VirusSampleFASTA(pg)

        sideEffect = Open(self, manager()).sideEffect
        with patch.object(builtins, 'open') as mockMethod:
            mockMethod.side_effect = sideEffect
            filename = virusSampleFASTA.add('Lausannevirus', 'filename-1')
            self.assertEqual('out/virus-0-sample-0.fasta', filename)
            self.assertEqual('>id1\nACTG\n', fastaIO.getvalue())

            # Repeated call. The side effect open will fail if open is
            # called at this point.
            filename = virusSampleFASTA.add('Lausannevirus', 'filename-1')
            self.assertEqual('out/virus-0-sample-0.fasta', filename)

    def testIdenticalReadsRemoved(self):
        """
        If two proteins in the same virus are matched by the same read, the
        de-duplicated FASTA for the virus must have only one copy of the
        duplicated read.
        """
        class Open(object):
            def __init__(self, test, manager):
                self.test = test
                self.manager = manager
                self.count = 0

            def sideEffect(self, filename, *args, **kwargs):
                if self.count == 0:
                    self.test.assertEqual('out/0.fasta', filename)
                    self.count += 1
                    return File(['>id1\n', 'ACTG\n'])
                elif self.count == 1:
                    self.test.assertEqual('out/1.fasta', filename)
                    self.count += 1
                    return File(['>id1\n', 'ACTG\n', '>id2\n', 'CAGT\n'])
                elif self.count == 2:
                    self.test.assertEqual('out/virus-0-sample-0.fasta',
                                          filename)
                    self.count += 1
                    return self.manager
                else:
                    self.test.fail(
                        'We are only supposed to be called 3 times. '
                        'Filename: %r, Args: %r, Keyword args: %r.' %
                        (filename, args, kwargs))

        fp = StringIO(
            '0.63 41.3 44.2 9 9 12 gi|327410| protein 77 [Lausannevirus]\n'
            '0.77 46.6 48.1 5 6 74 gi|327409| ubiquitin [Lausannevirus]\n'
        )
        fastaIO = StringIO()

        @contextmanager
        def manager():
            yield fastaIO

        pg = ProteinGrouper()
        pg.addFile('filename-1', fp)
        virusSampleFASTA = VirusSampleFASTA(pg)

        sideEffect = Open(self, manager()).sideEffect
        with patch.object(builtins, 'open') as mockMethod:
            mockMethod.side_effect = sideEffect
            filename = virusSampleFASTA.add('Lausannevirus', 'filename-1')
            self.assertEqual('out/virus-0-sample-0.fasta', filename)
            self.assertEqual('>id1\nACTG\n>id2\nCAGT\n', fastaIO.getvalue())



import numpy as np
from unittest import TestCase, skipUnless

try:
    from matplotlib import pyplot as plt
except ImportError:
    import platform
    if platform.python_implementation() == 'PyPy':
        havePlt = False
        # PyPy doesn't have a version of matplotlib. Make a fake class that
        # raises if used.

        class plt(object):
            def __getattr__(self, _):
                raise NotImplementedError(
                    'matplotlib is not supported under pypy')
    else:
        raise
else:
    havePlt = True

try:
    from unittest.mock import call, MagicMock, ANY
except ImportError:
    from mock import call, MagicMock, ANY

from Bio.SeqFeature import FeatureLocation, SeqFeature
from Bio.SeqRecord import SeqRecord

from dark.features import (Feature, FeatureList, _FeatureAdder,
                           ProteinFeatureAdder, NucleotideFeatureAdder)


class Test_Feature(TestCase):
    """
    Tests of the C{Feature} class.
    """

    def testNotSubfeatureByDefault(self):
        """
        A feature must not be a subfeature by default.
        """
        location = FeatureLocation(100, 200)
        seqFeature = SeqFeature(location=location)
        feature = Feature(seqFeature)
        self.assertFalse(feature.subfeature)

    def testSubfeature(self):
        """
        A feature must be a subfeature if it is passed C{subfeature} = C{True}
        on initialization.
        """
        location = FeatureLocation(100, 200)
        seqFeature = SeqFeature(location=location)
        feature = Feature(seqFeature, subfeature=True)
        self.assertTrue(feature.subfeature)

    def testStart(self):
        """
        The start attribute must hold the feature's start.
        """
        location = FeatureLocation(100, 200)
        seqFeature = SeqFeature(location=location)
        feature = Feature(seqFeature)
        self.assertEqual(100, feature.start)

    def testEnd(self):
        """
        The end attribute must hold the feature start.
        """
        location = FeatureLocation(100, 200)
        seqFeature = SeqFeature(location=location)
        feature = Feature(seqFeature)
        self.assertEqual(200, feature.end)

    def testSetColor(self):
        """
        The setColor method must set the 'color' attribute on the feature.
        """
        location = FeatureLocation(100, 200)
        seqFeature = SeqFeature(location=location)
        feature = Feature(seqFeature)
        feature.setColor('red')
        self.assertEqual('red', feature.color)

    def testLegendLabel(self):
        """
        The legendLabel method must return a description.
        """
        location = FeatureLocation(100, 200)
        qualifiers = {
            'note': ['Capsid protein'],
        }
        seqFeature = SeqFeature(location=location, type='site',
                                qualifiers=qualifiers)
        feature = Feature(seqFeature)
        self.assertEqual('100-200 site. note: Capsid protein',
                         feature.legendLabel())

    def testLegendLabelQualifiersSorted(self):
        """
        The legendLabel method must return a description with sorted
        qualifiers.
        """
        location = FeatureLocation(100, 200)
        qualifiers = {
            'note': ['Capsid protein'],
            'product': ['CP1'],
        }
        seqFeature = SeqFeature(location=location, type='site',
                                qualifiers=qualifiers)
        feature = Feature(seqFeature)
        self.assertEqual('100-200 site. note: Capsid protein, product: CP1',
                         feature.legendLabel())

    def testLegendLabelNoQualifiers(self):
        """
        The legendLabel method must return a correct description for a feature
        that has no qualifiers.
        """
        location = FeatureLocation(100, 200)
        seqFeature = SeqFeature(location=location, type='site')
        feature = Feature(seqFeature)
        self.assertEqual('100-200 site.', feature.legendLabel())

    def testLegendLabelTruncatesValues(self):
        """
        The legendLabel method must return a description with qualifier
        values truncated to length 30 (including the trailing ...).
        """
        location = FeatureLocation(100, 200)
        qualifiers = {
            'note': ['x' * 40],
        }
        seqFeature = SeqFeature(location=location, type='site',
                                qualifiers=qualifiers)
        feature = Feature(seqFeature)
        xs = 'x' * 27 + '...'
        self.assertEqual('100-200 site. note: %s' % xs,
                         feature.legendLabel())


# We can't test anything that uses a FeatureList under pypy because
# dark.features makes use of matplotlib.
@skipUnless(havePlt, 'matplotlib not supported under pypy')
class Test_FeatureList(TestCase):
    """
    Tests of the C{FeatureList} class.
    """

    def testOffline(self):
        """
        If the sequence fetcher returns C{None} we must be marked as being
        offline.
        """
        fetcher = lambda title, db='database': None
        featureList = FeatureList('title', 'database', set(),
                                  sequenceFetcher=fetcher)
        self.assertEqual(True, featureList.offline)

    def testOfflineLength(self):
        """
        If the sequence fetcher indicates we're offline, the feature list
        must have length zero.
        """
        fetcher = lambda title, db='database': None
        featureList = FeatureList('title', 'database', set(),
                                  sequenceFetcher=fetcher)
        self.assertEqual(0, len(featureList))

    def testFetcherValueError(self):
        """
        If the sequence fetcher throws a C{ValueError}, the feature list
        must have length zero and should not be marked as being offline.
        """
        def fetcher(title, db):
            raise ValueError()

        featureList = FeatureList('title', 'database', set(),
                                  sequenceFetcher=fetcher)
        self.assertEqual(0, len(featureList))
        self.assertEqual(False, featureList.offline)

    def testNotOffline(self):
        """
        If the sequence fetcher does not return C{None} we must be marked as
        being online.
        """
        fetcher = lambda title, db='database': SeqRecord(None)
        featureList = FeatureList('title', 'database', set(),
                                  sequenceFetcher=fetcher)
        self.assertEqual(False, featureList.offline)

    def testNoFeaturesLength(self):
        """
        If the sequence fetcher returns a record with no features, the
        L{FeatureList} instance must have length zero.
        """
        fetcher = lambda title, db='database': SeqRecord(None)
        featureList = FeatureList('title', 'database', set(),
                                  sequenceFetcher=fetcher)
        self.assertEqual(0, len(featureList))

    def testNoQualifiersLength(self):
        """
        If the sequence fetcher returns a record with two features but
        neither of them has any qualifiers, the L{FeatureList} instance
        must still include both features.
        """
        def fetcher(title, db='database'):
            location = FeatureLocation(100, 200)
            feature = SeqFeature(type='site', location=location)
            return SeqRecord(None, features=[feature, feature])

        featureList = FeatureList('title', 'database', set(['site']),
                                  sequenceFetcher=fetcher)
        self.assertEqual(2, len(featureList))

    def testNotSubfeature(self):
        """
        If the sequence fetcher returns a record with a feature that is
        not a subfeature, it must not be marked as a subfeature.
        """
        def fetcher(title, db='database'):
            location = FeatureLocation(100, 200)
            feature = SeqFeature(type='site', qualifiers={'a': ['b']},
                                 location=location)
            return SeqRecord(None, features=[feature])

        featureList = FeatureList('title', 'database', set(['site']),
                                  sequenceFetcher=fetcher)
        self.assertFalse(featureList[0].subfeature)

    def testWantedTypeLength(self):
        """
        If the sequence fetcher returns a record with two features but
        only one of them has a wanted type ('site'), the L{FeatureList}
        instance must have length one.
        """
        def fetcher(title, db='database'):
            location = FeatureLocation(100, 200)
            feature1 = SeqFeature(type='region', location=location)
            feature2 = SeqFeature(type='site', qualifiers={'a': ['b']},
                                  location=location)
            return SeqRecord(None, features=[feature1, feature2])

        featureList = FeatureList('title', 'database', set(['site']),
                                  sequenceFetcher=fetcher)
        self.assertEqual(1, len(featureList))

    def testSubfeatures(self):
        """
        If the sequence fetcher returns a record with a feature that
        has a subfeature, the L{FeatureList} instance must have length two
        and the second feature in the list must be a subfeature.
        """
        def fetcher(title, db='database'):
            location = FeatureLocation(100, 200)
            subfeature = SeqFeature(type='site', qualifiers={'a': ['b']},
                                    location=location)
            feature = SeqFeature(type='site', qualifiers={'a': ['b']},
                                 sub_features=[subfeature], location=location)
            return SeqRecord(None, features=[feature])

        featureList = FeatureList('title', 'database', set(['site']),
                                  sequenceFetcher=fetcher)
        self.assertEqual(2, len(featureList))
        self.assertTrue(featureList[1].subfeature)

    def testSubfeaturesWithoutParentFeature(self):
        """
        If the sequence fetcher returns a record with a feature that
        has a subfeature, and the subfeature is wanted but the parent feature
        is not, the L{FeatureList} instance must have length one
        and the feature in the list must be a subfeature.
        """
        def fetcher(title, db='database'):
            location = FeatureLocation(100, 200)
            subfeature = SeqFeature(type='region', qualifiers={'a': ['b']},
                                    location=location)
            feature = SeqFeature(type='site', qualifiers={'a': ['b']},
                                 sub_features=[subfeature], location=location)
            return SeqRecord(None, features=[feature])

        featureList = FeatureList('title', 'database', set(['region']),
                                  sequenceFetcher=fetcher)
        self.assertEqual(1, len(featureList))
        self.assertTrue(featureList[0].subfeature)

    @skipUnless(havePlt, 'matplotlib not supported under pypy')
    def testColors(self):
        """
        If the sequence fetcher returns a record with 3 features, each
        must be assigned a correct color.
        """
        def fetcher(title, db='database'):
            location = FeatureLocation(100, 200)
            feature = SeqFeature(type='site', qualifiers={'a': ['b']},
                                 location=location)
            return SeqRecord(None, features=[feature] * 3)

        featureList = FeatureList('title', 'database', set(['site']),
                                  sequenceFetcher=fetcher)
        colormap = plt.cm.coolwarm
        colors = [colormap(i) for i in np.linspace(0.0, 0.99, 3)]

        for i in range(3):
            self.assertEqual(colors[i], featureList[i].color)


class Test_FeatureAdder(TestCase):
    """
    Tests of the C{_FeatureAdder} class.
    """

    def testNotTooManyFeaturesByDefault(self):
        """
        A new instance of L{_FeatureAdder} must be marked on creation as not
        having too many features to plot.
        """
        featureAdder = _FeatureAdder()
        self.assertFalse(featureAdder.tooManyFeaturesToPlot)

    @skipUnless(havePlt, 'matplotlib not supported under pypy')
    def testTitle(self):
        """
        A L{_FeatureAdder} must set the title of its figure.
        """
        fetcher = lambda title, db='database': None
        featureAdder = _FeatureAdder()
        fig = plt.subplot(111)
        fig.set_title = MagicMock()
        featureAdder.add(fig, 'title', 0, 100, sequenceFetcher=fetcher)
        fig.set_title.assert_called_with('Target sequence features',
                                         fontsize=16)

    @skipUnless(havePlt, 'matplotlib not supported under pypy')
    def testYTicks(self):
        """
        A L{_FeatureAdder} must set the title of its figure.
        """
        fetcher = lambda title, db='database': None
        featureAdder = _FeatureAdder()
        fig = plt.subplot(111)
        fig.set_yticks = MagicMock()
        featureAdder.add(fig, 'title', 0, 100, sequenceFetcher=fetcher)
        fig.set_yticks.assert_called_with([])

    @skipUnless(havePlt, 'matplotlib not supported under pypy')
    def testOffline(self):
        """
        If the sequence fetcher used by a L{_FeatureAdder} returns C{None},
        the C{add} method of the L{_FeatureAdder} must return C{None}. The
        C{text} and C{axis} methods on the fig must both be called correctly.
        """
        fetcher = lambda title, db='database': None
        featureAdder = _FeatureAdder()
        fig = plt.subplot(111)
        fig.text = MagicMock()
        fig.axis = MagicMock()
        featureAdder.add(fig, 'title', 0, 300, sequenceFetcher=fetcher)
        fig.text.assert_called_with(100, 0,
                                    'You (or Genbank) appear to be offline.',
                                    fontsize=20)
        fig.axis.assert_called_with([0, 300, -1, 1])

    @skipUnless(havePlt, 'matplotlib not supported under pypy')
    def testNoFeatures(self):
        """
        If the sequence fetcher used by a L{_FeatureAdder} returns no features
        the C{text} and C{axis} methods on the fig must both be called
        correctly and the C{add} method must return C{[]}.
        """
        fetcher = lambda title, db='database': SeqRecord(None)
        featureAdder = _FeatureAdder()
        featureAdder.WANTED_TYPES = ('site',)
        fig = plt.subplot(111)
        fig.text = MagicMock()
        fig.axis = MagicMock()
        result = featureAdder.add(fig, 'title', 0, 300,
                                  sequenceFetcher=fetcher)
        fig.text.assert_called_with(0.5, 0.5,
                                    'No features found',
                                    horizontalalignment='center',
                                    verticalalignment='center',
                                    transform=ANY, fontsize=20)
        fig.axis.assert_called_with([0, 300, -1, 1])
        self.assertEqual([], result)

    @skipUnless(havePlt, 'matplotlib not supported under pypy')
    def testOneFeatureRaisesNotImplementedError(self):
        """
        If the sequence fetcher used by a L{_FeatureAdder} returns a feature,
        the C{add} method must raise C{NotImplementedError} because the
        C{_displayFeatures} method is not implemented.
        """
        def fetcher(title, db='database'):
            location = FeatureLocation(100, 200)
            subfeature = SeqFeature(type='region', qualifiers={'a': ['b']},
                                    location=location)
            feature = SeqFeature(type='site', qualifiers={'a': ['b']},
                                 sub_features=[subfeature], location=location)
            return SeqRecord(None, features=[feature])

        featureAdder = _FeatureAdder()
        featureAdder.WANTED_TYPES = ('site',)
        fig = plt.subplot(111)
        self.assertRaises(NotImplementedError, featureAdder.add, fig, 'title',
                          0, 300, sequenceFetcher=fetcher)

    @skipUnless(havePlt, 'matplotlib not supported under pypy')
    def testTooManyFeatures(self):
        """
        If the sequence fetcher used by a L{_FeatureAdder} returns too many
        features, the C{text} and C{axis} methods on the figure must be called
        correctly and the C{add} call must return the sequences.
        """
        def fetcher(title, db='database'):
            location = FeatureLocation(100, 200)
            feature = SeqFeature(type='site', qualifiers={'a': ['b']},
                                 location=location)
            return SeqRecord(None, features=[feature] * 100)

        featureAdder = _FeatureAdder()
        featureAdder.WANTED_TYPES = ('site',)
        fig = plt.subplot(111)
        fig.text = MagicMock()
        fig.axis = MagicMock()
        result = featureAdder.add(fig, 'title', 0, 300,
                                  sequenceFetcher=fetcher)
        fig.text.assert_called_with(0.5, 0.5,
                                    'Too many features to plot',
                                    horizontalalignment='center',
                                    verticalalignment='center',
                                    transform=ANY, fontsize=20)
        fig.axis.assert_called_with([0, 300, -1, 1])
        self.assertTrue(isinstance(result, FeatureList))
        self.assertEqual(100, len(result))


class TestProteinFeatureAdder(TestCase):
    """
    Tests of the C{ProteinFeatureAdder} class.
    """

    @skipUnless(havePlt, 'matplotlib not supported under pypy')
    def testUnwantedFeature(self):
        """
        If the sequence fetcher used by a L{_FeatureAdder} returns a feature
        whose type is not wanted, the figure's plot method must not be called
        and the C{add} method must return an empty feature list.
        """
        def fetcher(title, db='database'):
            location = FeatureLocation(100, 200)
            feature = SeqFeature(type='unwanted', qualifiers={'a': ['b']},
                                 location=location)
            return SeqRecord(None, features=[feature])

        featureAdder = ProteinFeatureAdder()
        fig = plt.subplot(111)
        fig.plot = MagicMock()
        result = featureAdder.add(fig, 'title', 0, 300,
                                  sequenceFetcher=fetcher)
        self.assertEqual([], fig.plot.call_args_list)
        self.assertEqual([], result)

    @skipUnless(havePlt, 'matplotlib not supported under pypy')
    def testOneFeature(self):
        """
        If the sequence fetcher used by a L{_FeatureAdder} returns a feature,
        the C{text} and C{axis} methods on the figure must be called correctly
        and the C{add} call must return the sequences.
        """
        def fetcher(title, db='database'):
            location = FeatureLocation(100, 200)
            feature = SeqFeature(type='Site', qualifiers={'a': ['b']},
                                 location=location)
            return SeqRecord(None, features=[feature])

        featureAdder = ProteinFeatureAdder()
        fig = plt.subplot(111)
        fig.plot = MagicMock()
        fig.axis = MagicMock()
        fig.legend = MagicMock()
        result = featureAdder.add(fig, 'title', 0, 300,
                                  sequenceFetcher=fetcher)
        fig.plot.assert_called_with(
            [100, 200], [-0.0, -0.0],
            color=(0.2298057, 0.298717966, 0.75368315299999999, 1.0),
            linewidth=2)
        fig.axis.assert_called_with([0, 300, -0.4, 0.2])
        fig.legend.assert_called_with(
            ['100-200 Site. a: b'], loc='lower center', shadow=True,
            bbox_to_anchor=(0.5, 1.4), ncol=2, fancybox=True)
        self.assertTrue(isinstance(result, FeatureList))
        self.assertEqual(1, len(result))


class TestNucleotideFeatureAdder(TestCase):
    """
    Tests of the C{NucleotideFeatureAdder} class.
    """

    @skipUnless(havePlt, 'matplotlib not supported under pypy')
    def testUnwantedFeature(self):
        """
        If the sequence fetcher used by a L{_FeatureAdder} returns a feature
        whose type is not wanted, the figure's plot method must not be called
        and the C{add} method must return an empty feature list.
        """
        def fetcher(title, db='database'):
            location = FeatureLocation(100, 200)
            feature = SeqFeature(type='unwanted', qualifiers={'a': ['b']},
                                 location=location)
            return SeqRecord(None, features=[feature])

        featureAdder = NucleotideFeatureAdder()
        fig = plt.subplot(111)
        fig.plot = MagicMock()
        result = featureAdder.add(fig, 'title', 0, 300,
                                  sequenceFetcher=fetcher)
        self.assertEqual([], fig.plot.call_args_list)
        self.assertEqual([], result)

    @skipUnless(havePlt, 'matplotlib not supported under pypy')
    def testOneFeature(self):
        """
        If the sequence fetcher used by a L{_FeatureAdder} returns a feature,
        the C{text} and C{axis} methods on the figure must be called correctly
        and the C{add} call must return the sequences.
        """
        def fetcher(title, db='database'):
            location = FeatureLocation(100, 200)
            feature = SeqFeature(type='CDS', qualifiers={'a': ['b']},
                                 location=location)
            return SeqRecord(None, features=[feature])

        featureAdder = NucleotideFeatureAdder()
        fig = plt.subplot(111)
        fig.plot = MagicMock()
        fig.axis = MagicMock()
        fig.legend = MagicMock()
        result = featureAdder.add(fig, 'title', 0, 300,
                                  sequenceFetcher=fetcher)
        fig.plot.assert_called_with(
            [100, 200], [1, 1],
            color=(0.2298057, 0.298717966, 0.75368315299999999, 1.0),
            linewidth=2)
        fig.axis.assert_called_with([0, 300, -0.5, 2.5])
        fig.legend.assert_called_with(
            ['100-200 CDS. a: b'], loc='lower center',
            shadow=True, ncol=2, fancybox=True,
            bbox_to_anchor=(0.5, 2.5))
        self.assertTrue(isinstance(result, FeatureList))
        self.assertEqual(1, len(result))

    @skipUnless(havePlt, 'matplotlib not supported under pypy')
    def testOneFeatureAdjusted(self):
        """
        If the sequence fetcher used by a L{_FeatureAdder} returns a feature,
        the C{text} and C{axis} methods on the figure must be called correctly
        and the C{add} call must return the sequences.

        Note that offsets in the legend of nucleotide plots are adjusted. They
        shouldn't be as the adjusted offsets make no sense to the reader.
        """
        def fetcher(title, db='database'):
            location = FeatureLocation(100, 200)
            feature = SeqFeature(type='CDS', qualifiers={'a': ['b']},
                                 location=location)
            return SeqRecord(None, features=[feature])

        featureAdder = NucleotideFeatureAdder()
        fig = plt.subplot(111)
        fig.plot = MagicMock()
        fig.axis = MagicMock()
        fig.legend = MagicMock()
        adjuster = lambda x: 3 * x
        result = featureAdder.add(fig, 'title', 0, 300, adjuster,
                                  sequenceFetcher=fetcher)
        fig.plot.assert_called_with(
            [300, 600], [0, 0],
            color=(0.2298057, 0.298717966, 0.75368315299999999, 1.0),
            linewidth=2)
        fig.axis.assert_called_with([0, 300, -0.5, 2.5])
        fig.legend.assert_called_with(
            ['100-200 CDS. a: b'], loc='lower center',
            shadow=True, ncol=2, fancybox=True,
            bbox_to_anchor=(0.5, 2.5))
        self.assertTrue(isinstance(result, FeatureList))
        self.assertEqual(1, len(result))

    @skipUnless(havePlt, 'matplotlib not supported under pypy')
    def testSubfeaturesAreMovedDown(self):
        """
        If the sequence fetcher used by a L{_FeatureAdder} returns a feature,
        and a subfeature on the same frame, the subfeature must be plotted
        a little (0.2) below the feature.
        """
        def fetcher(title, db='database'):
            subfeature = SeqFeature(type='CDS', qualifiers={'a': ['b']},
                                    location=FeatureLocation(130, 150))
            feature = SeqFeature(type='CDS', qualifiers={'a': ['b']},
                                 location=FeatureLocation(100, 200),
                                 sub_features=[subfeature])
            return SeqRecord(None, features=[feature])

        featureAdder = NucleotideFeatureAdder()
        fig = plt.subplot(111)
        fig.plot = MagicMock()
        fig.axis = MagicMock()
        fig.legend = MagicMock()
        result = featureAdder.add(fig, 'title', 0, 300,
                                  sequenceFetcher=fetcher)
        self.assertEqual(
            fig.plot.call_args_list,
            [
                call([100, 200], [1, 1], color=ANY, linewidth=2),
                call([130, 150], [0.8, 0.8], color=ANY, linewidth=2),
            ])
        fig.axis.assert_called_with([0, 300, -0.5, 2.5])
        fig.legend.assert_called_with(
            ['100-200 CDS. a: b',
             '130-150 CDS (subfeature). a: b'],
            loc='lower center', shadow=True, ncol=2, fancybox=True,
            bbox_to_anchor=(0.5, 2.5))
        self.assertTrue(isinstance(result, FeatureList))
        self.assertEqual(2, len(result))

    @skipUnless(havePlt, 'matplotlib not supported under pypy')
    def testPolyproteinsAreMovedUp(self):
        """
        If the sequence fetcher used by a L{_FeatureAdder} returns a feature,
        that's a polyprotein, the feature must be plotted a little (0.2) above
        its normal location.
        """
        def fetcher(title, db='database'):
            feature1 = SeqFeature(type='CDS',
                                  qualifiers={'product': ['a polyprotein']},
                                  location=FeatureLocation(100, 200))
            feature2 = SeqFeature(type='CDS',
                                  qualifiers={'a': ['b']},
                                  location=FeatureLocation(130, 150))
            return SeqRecord(None, features=[feature1, feature2])

        featureAdder = NucleotideFeatureAdder()
        fig = plt.subplot(111)
        fig.plot = MagicMock()
        fig.axis = MagicMock()
        fig.legend = MagicMock()
        result = featureAdder.add(fig, 'title', 0, 300,
                                  sequenceFetcher=fetcher)
        self.assertEqual(
            fig.plot.call_args_list,
            [
                call([100, 200], [1.2, 1.2], color=ANY, linewidth=2),
                call([130, 150], [1.0, 1.0], color=ANY, linewidth=2),
            ])
        fig.axis.assert_called_with([0, 300, -0.5, 2.5])
        fig.legend.assert_called_with(
            ['100-200 CDS. product: a polyprotein',
             '130-150 CDS. a: b'],
            loc='lower center', shadow=True, ncol=2, fancybox=True,
            bbox_to_anchor=(0.5, 2.5))
        self.assertTrue(isinstance(result, FeatureList))
        self.assertEqual(2, len(result))



#!/usr/bin/env python

from __future__ import print_function

import sys
from Bio import SeqIO
from Bio.Seq import Seq
from Bio.Alphabet import IUPAC
from dark.sequence import findPrimerBidiLimits


def trimPrimers(primer, verbose):
    """
    @param primer: A BioPython C{Bio.Seq} primer sequence.
    @param verbose: A C{bool}, if C{True} output additional information about
        how often and where primers were found.
    """
    reads = []
    absentCount = forwardCount = reverseCount = count = 0
    for seqRecord in SeqIO.parse(sys.stdin, 'fasta'):
        count += 1
        start, end = findPrimerBidiLimits(primer, seqRecord.seq)
        if start == 0:
            if end == len(seqRecord):
                absentCount += 1
            else:
                reverseCount += 1
        else:
            forwardCount += 1
            if end != len(seqRecord):
                reverseCount += 1
        reads.append(seqRecord[start:end])

    if verbose:
        print((
            'Read %d sequences. Found forward: %d, '
            'Found reversed: %d, Absent: %d') % (
            count, forwardCount, reverseCount, absentCount), file=sys.stderr)

    SeqIO.write(reads, sys.stdout, 'fasta')


if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser(
        description=('Given FASTA on stdin, look for a primer sequence '
                     'and write trimmed FASTA (after the primer) to stdout.'))

    parser.add_argument('primer', help='the primer sequence')
    parser.add_argument('--verbose', type=bool, default=False,
                        help='If True, print information on found primers')

    args = parser.parse_args()

    trimPrimers(Seq(args.primer.upper(), IUPAC.unambiguous_dna),
                args.verbose)



from __future__ import division, print_function

import re
from os.path import dirname, join
from operator import itemgetter
from six.moves.urllib.parse import quote
import numpy as np
from textwrap import fill
from math import log10

try:
    import matplotlib.pyplot as plt
except ImportError:
    import platform
    if platform.python_implementation() == 'PyPy':
        raise NotImplementedError(
            'matplotlib is not supported under pypy')
    else:
        raise

from dark.dimension import dimensionalIterator
from dark.fasta import FastaReads
from dark.html import NCBISequenceLinkURL
from dark.reads import Reads


def readSampleTotalReadCounts(fp):
    """
    Read a file of total sample read counts and return it as a C{dict}.

    @param fp: An open file pointer to read from.
    @raise ValueError: If a count is not an integer.
    @return: A C{dict} mapping C{str} sample names to C{int} sample total read
        count.
    """
    result = {}
    for line in fp:
        # The following implicitly assumes sample names do not have a
        # space. We could change that by putting the count first in the
        # file or by finding the last whitespace and splitting after that
        # to get the count. For now I'm just noting the assumption.
        sampleName, count = line.split(None, 1)
        result[sampleName] = int(count)
    return result


class VirusSampleFASTA(object):
    """
    Maintain a cache of virus/sample FASTA file names, creating de-duplicaed
    FASTA files (from reads for all proteins of a virus that a sample has)
    on demand.

    @param proteinGrouper: An instance of C{ProteinGrouper}.
    """
    def __init__(self, proteinGrouper):
        self._proteinGrouper = proteinGrouper
        self._viruses = {}
        self._samples = {}
        self._fastaFilenames = {}

    def add(self, virusTitle, sampleName):
        """
        Add a virus title, sample name combination and get its FASTA file
        name and unique read count. Write the FASTA file if it does not
        already exist. Save the unique read count into
        C{self._proteinGrouper}.

        @param virusTitle: A C{str} virus title.
        @param sampleName: A C{str} sample name.
        @return: A C{str} giving the FASTA file name holding all the reads
            (without duplicates) from the sample that matched the proteins in
            the given virus.
        """
        virusIndex = self._viruses.setdefault(virusTitle, len(self._viruses))
        sampleIndex = self._samples.setdefault(sampleName, len(self._samples))

        try:
            return self._fastaFilenames[(virusIndex, sampleIndex)]
        except KeyError:
            reads = Reads()
            for protein in self._proteinGrouper.virusTitles[
                    virusTitle][sampleName]['proteins']:
                for read in FastaReads(protein['fastaFilename'],
                                       checkAlphabet=0):
                    reads.add(read)
            saveFilename = join(
                protein['outDir'],
                'virus-%d-sample-%d.fasta' % (virusIndex, sampleIndex))
            unique = reads.filter(removeDuplicates=True)
            unique.save(saveFilename)
            # Save the unique count into self._proteinGrouper
            self._proteinGrouper.virusTitles[
                virusTitle][sampleName]['uniqueReadCount'] = len(unique)
            self._fastaFilenames[(virusIndex, sampleIndex)] = saveFilename
            return saveFilename

    def lookup(self, virusTitle, sampleName):
        """
        Look up a virus title, sample name combination and get its FASTA file
        name and unique read count.

        This method should be used instead of C{add} in situations where
        you want an exception to be raised if a virus/sample combination has
        not already been passed to C{add}.

        @param virusTitle: A C{str} virus title.
        @param sampleName: A C{str} sample name.
        @raise KeyError: If the virus title or sample name have not been seen,
            either individually or in combination.
        @return: A (C{str}, C{int}) tuple retrieved from self._fastaFilenames
        """
        virusIndex = self._viruses[virusTitle]
        sampleIndex = self._samples[sampleName]
        return self._fastaFilenames[(virusIndex, sampleIndex)]


class ProteinGrouper(object):
    """
    Group matched proteins by the virus they come from.

    @param assetDir: The C{str} directory name where
        C{noninteractive-alignment-panel.py} put its HTML, blue plot and
        alignment panel images, and FASTA files. This must be relative
        to the filenames that will later be passed to C{addFile}.
    @param sampleNameRegex: A C{str} regular expression that can be used to
        extract a short sample name from full file names subsequently passed
        to C{self.addFile}. The regular expression must have a matching group
        (delimited by parentheses) to capture the part of the file name that
        should be used as the sample name.
    """

    # The following regex is deliberately greedy (using .*) to consume the
    # whole protein title before backtracking to find the last [virus title]
    # section. That way, it will match just the last [virus title] in a
    # protein. This avoids situations in which two [...] delimited substrings
    # are present in a protein name (in which case we just want the last).
    # E.g., the following is a complete protein name:
    #
    #   gi|19919894|ref|NP_612577.1| Enzymatic polyprotein [Contains: Aspartic
    #   protease; Endonuclease; Reverse transcriptase] [Carnation etched ring
    #   virus]
    #
    # Unfortunately the regex doesn't find the vius title when the protein
    # title has nested [...] sections, as in this example:
    #
    #   gi|224808893|ref|YP_002643049.1| replication-associated protein [Tomato
    #   leaf curl Nigeria virus-[Nigeria:2006]]
    #
    # I decided not to worry about nested [...] sections (there are only 2
    # instances that I know of).
    VIRUS_RE = re.compile('^(.*)\[([^\]]+)\]$')

    # The virus title assigned to proteins whose title strings cannot be parsed
    # for a virus title (see previous comment).  Do not use '<', '>' or any
    # other HTML special chars in the following.
    NO_VIRUS_TITLE = '[no virus name found in protein title]'

    VIRALZONE = 'http://viralzone.expasy.org/cgi-bin/viralzone/search?query='

    def __init__(self, assetDir='out', sampleNameRegex=None):
        self._sampleNameRegex = (re.compile(sampleNameRegex) if sampleNameRegex
                                 else None)
        self._assetDir = assetDir
        # virusTitles will be a dict of dicts of dicts. The first two keys
        # will be a virus title and a sample name. The final dict will
        # contain 'proteins' (a list of dicts) and 'uniqueReadCount' (an int).
        self.virusTitles = {}
        # sampleNames is keyed by sample name and will have values that hold
        # the sample's alignment panel index.html file.
        self.sampleNames = {}
        self.virusSampleFASTA = VirusSampleFASTA(self)

    def _title(self):
        """
        Create a title summarizing the viruses and samples.

        @return: A C{str} title.
        """
        return (
            '%d virus%s found in %d sample%s' %
            (len(self.virusTitles), '' if len(self.virusTitles) == 1 else 'es',
             len(self.sampleNames), '' if len(self.sampleNames) == 1 else 's'))

    def addFile(self, filename, fp):
        """
        Read and record protein information for a sample.

        @param filename: A C{str} file name.
        @param fp: An open file pointer to read the file's data from.
        """

        if self._sampleNameRegex:
            match = self._sampleNameRegex.search(filename)
            if match:
                sampleName = match.group(1)
            else:
                sampleName = filename
        else:
            sampleName = filename

        outDir = join(dirname(filename), self._assetDir)

        self.sampleNames[sampleName] = join(outDir, 'index.html')

        for index, proteinLine in enumerate(fp):
            proteinLine = proteinLine[:-1]
            try:
                (coverage, medianScore, bestScore, readCount, hspCount,
                 proteinLength, titles) = proteinLine.split(None, 6)
            except ValueError as e:
                raise ValueError(
                    'Could not split line %d of file %r into 7 fields. '
                    'Line was %r. ValueError from split: %s.' %
                    (index + 1, filename, proteinLine, e))

            match = self.VIRUS_RE.match(titles)
            if match:
                proteinTitle = match.group(1).strip()
                virusTitle = match.group(2)
            else:
                proteinTitle = titles
                virusTitle = self.NO_VIRUS_TITLE

            if virusTitle not in self.virusTitles:
                self.virusTitles[virusTitle] = {}

            if sampleName not in self.virusTitles[virusTitle]:
                self.virusTitles[virusTitle][sampleName] = {
                    'proteins': [],
                    'uniqueReadCount': None,
                }

            self.virusTitles[virusTitle][sampleName]['proteins'].append({
                'bestScore': float(bestScore),
                'bluePlotFilename': join(outDir, '%d.png' % index),
                'coverage': float(coverage),
                'fastaFilename': join(outDir, '%d.fasta' % index),
                'hspCount': int(hspCount),
                'index': index,
                'medianScore': float(medianScore),
                'outDir': outDir,
                'proteinLength': int(proteinLength),
                'proteinTitle': proteinTitle,
                'proteinURL': NCBISequenceLinkURL(proteinTitle),
                'readCount': int(readCount),
            })

    def _computeUniqueReadCounts(self):
        """
        Add all virus / sample combinations to self.virusSampleFASTA.

        This will make all de-duplicated FASTA files and store the number
        of de-duplicated reads into C{self.virusTitles}.
        """
        for virusTitle, samples in self.virusTitles.items():
            for sampleName in samples:
                self.virusSampleFASTA.add(virusTitle, sampleName)

    def toStr(self):
        """
        Produce a string representation of the virus summary.

        @return: A C{str} suitable for printing.
        """
        # Note that the string representation contains much less
        # information than the HTML summary. E.g., it does not contain the
        # unique (de-duplicated) read count, since that is only computed
        # when we are making combined FASTA files of reads matching a
        # virus.
        titleGetter = itemgetter('proteinTitle')
        readCountGetter = itemgetter('readCount')
        result = []
        append = result.append

        append(self._title())
        append('')

        for virusTitle in sorted(self.virusTitles):
            samples = self.virusTitles[virusTitle]
            sampleCount = len(samples)
            append('%s (in %d sample%s)' %
                   (virusTitle,
                    sampleCount, '' if sampleCount == 1 else 's'))
            for sampleName in sorted(samples):
                proteins = samples[sampleName]['proteins']
                proteinCount = len(proteins)
                totalReads = sum(readCountGetter(p) for p in proteins)
                append('  %s (%d protein%s, %d read%s)' %
                       (sampleName,
                        proteinCount, '' if proteinCount == 1 else 's',
                        totalReads, '' if totalReads == 1 else 's'))
                proteins.sort(key=titleGetter)
                for proteinMatch in proteins:
                    append(
                        '    %(coverage).2f\t%(medianScore).2f\t'
                        '%(bestScore).2f\t%(readCount)4d\t%(hspCount)4d\t'
                        '%(index)3d\t%(proteinTitle)s'
                        % proteinMatch)
            append('')

        return '\n'.join(result)

    def toHTML(self, virusPanelFilename=None,
               sampleTotalReadCountFilename=None):
        """
        Produce an HTML string representation of the virus summary.

        @param virusPanelFilename: If not C{None}, a C{str} filename to write
            a virus panel PNG image to.
        @param sampleTotalReadCountFilename: If not C{None}, a C{str} filename
            providing total read count per sample. The file must contain lines
            with a sample name, whitespace, then a non-negative integer read
            count.
        @return: An HTML C{str} suitable for printing.
        """
        self._computeUniqueReadCounts()

        if virusPanelFilename:
            self.virusPanel(virusPanelFilename, sampleTotalReadCountFilename)

        titleGetter = itemgetter('proteinTitle')
        virusTitles = sorted(self.virusTitles)
        sampleNames = sorted(self.sampleNames)

        result = [
            '<html>',
            '<head>',
            '<title>',
            self._title(),
            '</title>',
            '</head>',
            '<body>',
            '<style>',
            '''\
            body {
                margin-left: 2%;
                margin-right: 2%;
            }
            .sample {
                margin-bottom: 2px;
            }
            .sample-name {
                color: red;
            }
            .index {
                font-size: small;
            }
            .protein-title {
                font-family: "Courier New", Courier, monospace;
            }
            .stats {
                font-family: "Courier New", Courier, monospace;
                white-space: pre;
            }
            .protein-list {
                margin-top: 2px;
            }''',
            '</style>',
            '</head>',
            '<body>',
        ]

        proteinFieldsDescription = (
            '<p>',
            'In the bullet point protein lists below, there are eight fields:',
            '<ol>',
            '<li>Coverage fraction.</li>',
            '<li>Median bit score.</li>',
            '<li>Best bit score.</li>',
            '<li>Read count.</li>',
            '<li>HSP count (a read can match a protein more than once).</li>',
            '<li>Protein length (in AAs).</li>',
            '<li>Index (just ignore this).</li>',
            '<li>Protein name.</li>',
            '</ol>',
            '</p>',
        )

        append = result.append

        append('<h1>%s</h1>' % self._title())

        if virusPanelFilename:
            append('<p>')
            append('<a href="%s">Panel showing read count per virus, per '
                   'sample.</a>' % virusPanelFilename)
            append('Red vertical bars indicate samples with an unusually high '
                   'read count.')
            append('</p>')

        # Write a linked table of contents by virus.
        append('<h2>Virus index</h2>')
        append('<p class="index">')
        for virusTitle in virusTitles:
            append('<a href="#virus-%s">%s</a>' % (virusTitle, virusTitle))
            append('&middot;')
        # Get rid of final middle dot.
        result.pop()
        append('</p>')

        # Write a linked table of contents by sample.
        append('<h2>Sample index</h2>')
        append('<p class="index">')
        for sampleName in sampleNames:
            append('<a href="#sample-%s">%s</a>' % (sampleName, sampleName))
            append('&middot;')
        # Get rid of final middle dot.
        result.pop()
        append('</p>')

        # Write all viruses (with samples (with proteins)).
        append('<h1>Viruses by sample</h1>')
        result.extend(proteinFieldsDescription)

        for virusTitle in virusTitles:
            samples = self.virusTitles[virusTitle]
            sampleCount = len(samples)
            append(
                '<a id="virus-%s">'
                '<h2 class="virus"><span class="virus-title">%s</span> '
                '(in %d sample%s, <a href="%s%s">viralzone</a>)</h2>' %
                (virusTitle, virusTitle, sampleCount,
                 '' if sampleCount == 1 else 's',
                 self.VIRALZONE, quote(virusTitle)))
            for sampleName in sorted(samples):
                fastaName = self.virusSampleFASTA.lookup(virusTitle,
                                                         sampleName)
                proteins = samples[sampleName]['proteins']
                proteinCount = len(proteins)
                uniqueReadCount = samples[sampleName]['uniqueReadCount']
                append(
                    '<p class=sample>'
                    '<span class="sample-name">%s</span> '
                    '(%d protein%s, %d read%s, <a href="%s">panel</a>, '
                    '<a href="%s">fasta</a>)' %
                    (sampleName,
                     proteinCount, '' if proteinCount == 1 else 's',
                     uniqueReadCount, '' if uniqueReadCount == 1 else 's',
                     self.sampleNames[sampleName], fastaName))
                proteins.sort(key=titleGetter)
                append('<ul class="protein-list">')
                for proteinMatch in proteins:
                    append(
                        '<li>'
                        '<span class="stats">'
                        '%(coverage).2f %(medianScore)6.2f %(bestScore)6.2f '
                        '%(readCount)4d %(hspCount)4d %(proteinLength)4d '
                        '%(index)3d '
                        '</span> '
                        '<span class="protein-title">'
                        '%(proteinTitle)s'
                        '</span> '
                        '(<a href="%(bluePlotFilename)s">blue plot</a>, '
                        '<a href="%(fastaFilename)s">fasta</a>'
                        % proteinMatch)

                    if proteinMatch['proteinURL']:
                        # Append this directly to the last string in result, to
                        # avoid introducing whitespace when we join result
                        # using '\n'.
                        result[-1] += (', <a href="%s">NCBI</a>' %
                                       proteinMatch['proteinURL'])
                    result[-1] += ')'

                    append('</li>')

                append('</ul>')
                append('</p>')

        # Write all samples (with viruses (with proteins)).
        append('<h1>Samples by virus</h1>')
        result.extend(proteinFieldsDescription)

        for sampleName in sampleNames:
            sampleVirusTitles = set()
            for virusTitle in virusTitles:
                if sampleName in self.virusTitles[virusTitle]:
                    sampleVirusTitles.add(virusTitle)

            append(
                '<a id="sample-%s">'
                '<h2 class="sample"><span class="sample-name">%s</span> '
                '(has proteins from %d virus%s, <a href="%s">panel</a>)</h2>' %
                (sampleName, sampleName, len(sampleVirusTitles),
                 '' if len(sampleVirusTitles) == 1 else 'es',
                 self.sampleNames[sampleName]))

            for virusTitle in sorted(sampleVirusTitles):
                fastaName = self.virusSampleFASTA.lookup(virusTitle,
                                                         sampleName)
                proteins = self.virusTitles[virusTitle][sampleName]['proteins']
                uniqueReadCount = self.virusTitles[
                    virusTitle][sampleName]['uniqueReadCount']
                proteinCount = len(proteins)
                append(
                    '<p class="sample">'
                    '<span class="virus-title">%s</span> '
                    '(%d protein%s, %d read%s, <a href="%s">fasta</a>)' %
                    (virusTitle,
                     proteinCount, '' if proteinCount == 1 else 's',
                     uniqueReadCount, '' if uniqueReadCount == 1 else 's',
                     fastaName))
                proteins.sort(key=titleGetter)
                append('<ul class="protein-list">')
                for proteinMatch in proteins:
                    append(
                        '<li>'
                        '<span class="stats">'
                        '%(coverage).2f %(medianScore)6.2f %(bestScore)6.2f '
                        '%(readCount)4d %(hspCount)4d %(proteinLength)4d '
                        '%(index)3d '
                        '</span> '
                        '<span class="protein-title">'
                        '%(proteinTitle)s'
                        '</span> '
                        '(<a href="%(bluePlotFilename)s">blue plot</a>, '
                        '<a href="%(fastaFilename)s">fasta</a>'
                        % proteinMatch)

                    if proteinMatch['proteinURL']:
                        # Append this directly to the last string in result, to
                        # avoid introducing whitespace when we join result
                        # using '\n'.
                        result[-1] += (', <a href="%s">NCBI</a>' %
                                       proteinMatch['proteinURL'])
                    result[-1] += ')'

                    append('</li>')

                append('</ul>')
                append('</p>')

        append('</body>')
        append('</html>')

        return '\n'.join(result)

    @staticmethod
    def _normalizedReadCountColors(readCounts, sampleNames):
        """
        Assign a color to each sample based on its normalized read count.

        @param readCounts: A C{list} of C{int} read counts, corresponding
            to the sample names in C{sampleNames}.
        @param sampleNames: A sorted C{list} of sample names.
        @return: A 2-tuple of (color, highlightedSamples) where 'color' is
            a C{list} of colors for samples and 'highlightedSamples' is a
            C{list} of names of samples that were assigned the highlight color.
        """
        highlight = 'r'
        normal = 'gray'
        minLoggedReadsForHighlighting = -6.0
        highlightedSamples = []
        color = []

        for readCount, sampleName in zip(readCounts, sampleNames):
            if readCount > minLoggedReadsForHighlighting:
                color.append(highlight)
                highlightedSamples.append(sampleName)
            else:
                color.append(normal)

        return color, highlightedSamples

    @staticmethod
    def _readCountColors(readCounts, sampleNames):
        """
        Assign a color to each sample based on its read count.

        @param readCounts: A C{list} of C{int} read counts, corresponding
            to the sample names in C{sampleNames}.
        @param sampleNames: A sorted C{list} of sample names.
        @return: A 2-tuple of (color, highlightedSamples) where 'color' is
            a C{list} of colors for samples and 'highlightedSamples' is a
            C{list} of names of samples that were assigned the highlight color.
        """
        highlight = 'r'
        normal = 'gray'
        sdMultiple = 2.5
        minReadsForHighlighting = 10
        highlightedSamples = []

        if len(readCounts) == 1:
            if readCounts[0] > minReadsForHighlighting:
                color = [highlight]
                highlightedSamples.append(sampleNames[0])
            else:
                color = [normal]
        else:
            mean = np.mean(readCounts)
            sd = np.std(readCounts)
            color = []
            for readCount, sampleName in zip(readCounts, sampleNames):
                if (readCount > (sdMultiple * sd) + mean and
                        readCount >= minReadsForHighlighting):
                    color.append(highlight)
                    highlightedSamples.append(sampleName)
                else:
                    color.append(normal)

        return color, highlightedSamples

    def _virusSamplePlot(self, virusTitle, sampleNames, ax,
                         sampleTotalReadCounts):
        """
        Make a plot of a graph with sample read count for a virus (Y axis)
        versus sample id (X axis).

        @param virusTitle: A C{str} virus title.
        @param sampleNames: A sorted C{list} of sample names.
        @param ax: A matplotlib C{axes} instance.
        @param sampleTotalReadCounts: Either a C{dict} mapping sample names to
            C{int} total read count for the sample, or C{None} if the total
            read counts are unknown.
        """
        LOW_LOG = -10.0
        readCounts = []
        for i, sampleName in enumerate(sampleNames):
            try:
                readCount = self.virusTitles[virusTitle][sampleName][
                    'uniqueReadCount']
            except KeyError:
                readCount = 0

            if sampleTotalReadCounts:
                ax.set_ylim([LOW_LOG, 1.0])
                try:
                    fraction = readCount / sampleTotalReadCounts[sampleName]
                except ZeroDivisionError:
                    raise ValueError('Sample %r has a zero total read count' %
                                     sampleName)
                else:
                    if fraction == 0:
                        readCounts.append(LOW_LOG)
                    else:
                        try:
                            loggedFraction = log10(fraction)
                        except ValueError as e:
                            raise ValueError(
                                'Sample %r read count %d total reads %d '
                                'fraction %.20f got a ValueError: %s'
                                % (sampleName, readCount,
                                   sampleTotalReadCounts[sampleName],
                                   fraction, e))
                        else:
                            readCounts.append(loggedFraction)
            else:
                readCounts.append(readCount)

        if sampleTotalReadCounts:
            color, highlightedSamples = self._normalizedReadCountColors(
                readCounts, sampleNames)
        else:
            color, highlightedSamples = self._readCountColors(
                readCounts, sampleNames)

        nSamples = len(sampleNames)
        x = np.arange(nSamples)
        ax.set_xticks([])
        ax.set_xlim((-0.5, nSamples - 0.5))

        if sampleTotalReadCounts:
            yMin = [LOW_LOG] * nSamples
        else:
            yMin = np.zeros(nSamples)
        ax.vlines(x, yMin, readCounts, color=color)

        if highlightedSamples:
            title = '%s\nIn red: %s' % (
                virusTitle, fill(', '.join(highlightedSamples), 50))
        else:
            # Add a newline to keep the first line of each title at the
            # same place as those titles that have an "In red:" second
            # line.
            title = virusTitle + '\n'

        ax.set_title(title, fontsize=10)
        ax.tick_params(axis='both', which='major', labelsize=8)
        ax.tick_params(axis='both', which='minor', labelsize=6)

    def virusPanel(self, filename, sampleTotalReadCountFilename=None):
        """
        Make a panel of images, with each image being a graph giving virus
        de-duplicated read count (Y axis) versus sample id (X axis).

        @param filename: A C{str} file name to write the image to.
        @param sampleTotalReadCountFilename: If not C{None}, a C{str} filename
            providing total read count per sample. The file must contain lines
            with a sample name, whitespace, then a non-negative integer read
            count.
        @raise IOError: If C{sampleTotalReadCountFilename} names a non-existent
            file.
        """
        self._computeUniqueReadCounts()
        virusTitles = sorted(self.virusTitles)
        sampleNames = sorted(self.sampleNames)

        if sampleTotalReadCountFilename:
            with open(sampleTotalReadCountFilename) as fp:
                sampleTotalReadCounts = readSampleTotalReadCounts(fp)
        else:
            sampleTotalReadCounts = None

        cols = 5
        rows = int(len(virusTitles) / cols) + (
            0 if len(virusTitles) % cols == 0 else 1)
        figure, ax = plt.subplots(rows, cols)  # , squeeze=False)

        coords = dimensionalIterator((rows, cols))

        for i, virusTitle in enumerate(virusTitles):
            row, col = next(coords)
            self._virusSamplePlot(virusTitle, sampleNames, ax[row][col],
                                  sampleTotalReadCounts=sampleTotalReadCounts)

        # Hide the final panel graphs (if any) that have no content. We do
        # this because the panel is a rectangular grid and some of the
        # plots at the end of the last row may be unused.
        for row, col in coords:
            ax[row][col].axis('off')

        figure.suptitle(
            ('Per-sample read count for %d virus%s and %d sample%s.\n\nSample '
             'name%s: %s') % (
                 len(virusTitles),
                 '' if len(virusTitles) == 1 else 'es',
                 len(sampleNames),
                 '' if len(sampleNames) == 1 else 's',
                 '' if len(sampleNames) == 1 else 's',
                 fill(', '.join(sampleNames), 50)),
            fontsize=20)
        figure.set_size_inches(5.0 * cols, 2.0 * rows, forward=True)
        plt.subplots_adjust(hspace=0.4)

        figure.savefig(filename)
